/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilegui;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.util.Hashtable;

/**
 * Centers all components in the middle of the parent container. The components will
 * be sized to their preferred size.
 * If the containers has more than one child, the children are painted on top of each other.
 * This layout manager is especially useful for centering a panel within other panel.
 * 
 * @author Michael Hedenus
 */
public class CenterLayout implements LayoutManager
{
	private Hashtable components = new Hashtable();

	public CenterLayout()
	{
		//NOP
	}

	public void addLayoutComponent(String constraint, Component comp)
	{
		components.put(comp, constraint);
	}

	public void layoutContainer(Container parent)
	{
		synchronized (parent.getTreeLock())
		{
			Insets insets = parent.getInsets();
			int x0 = ((parent.getWidth() - insets.left - insets.right) / 2) + insets.left;
			int y0 = ((parent.getHeight() - insets.top - insets.bottom) / 2) + insets.top;

			for (int i=0; i<parent.getComponentCount(); i++)
			{
				Component comp = parent.getComponent(i);
				Dimension prefSize = comp.getPreferredSize();
				comp.setBounds(x0 - prefSize.width / 2, y0 - prefSize.height / 2, prefSize.width, prefSize.height);
			}
		}
	}

	public Dimension minimumLayoutSize(Container parent)
	{
		synchronized (parent.getTreeLock())
		{
			Dimension minSize = new Dimension();
			for (int i=0; i<parent.getComponentCount(); i++)
			{
				Component comp = parent.getComponent(i);
				Dimension d = comp.getPreferredSize();
				if (d != null)
				{
					if (d.width > minSize.width)
					{
						minSize.width = d.width;
					}
					if (d.height > minSize.height)
					{
						minSize.height = d.height;
					}
				}
			}

			return minSize;
		}
	}

	public Dimension preferredLayoutSize(Container parent)
	{
		synchronized (parent.getTreeLock())
		{
			Dimension prefSize = new Dimension();

			for (int i=0; i<parent.getComponentCount(); i++)
			{
				Component comp = parent.getComponent(i);
				Dimension d = comp.getPreferredSize();
				if (d != null)
				{
					if (d.width > prefSize.width)
					{
						prefSize.width = d.width;
					}
					if (d.height > prefSize.height)
					{
						prefSize.height = d.height;
					}
				}
			}

			Insets insets = parent.getInsets();
			prefSize.width += (insets.left + insets.right);
			prefSize.height += (insets.top + insets.bottom);

			return prefSize;
		}
	}

	public void removeLayoutComponent(Component comp)
	{
		components.remove(comp);
	}
}
