/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilegui;

import java.awt.Component;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.Toolkit;
import java.net.URL;

/**
 * An icon that can be rendered.
 * 
 * @author Michael Hedenus
 */
public class Icon
{
	/**
	 * Id used in loading images from MediaTracker.
	 */
	private static int mediaTrackerID;

	private final static Component component = new Component()
	{
		private static final long serialVersionUID = 2069487492522466152L;
		//NOP
	};

	private final static MediaTracker tracker = new MediaTracker(component);

	/**
	 * Returns an ID to use with the MediaTracker in loading an image.
	 */
	private int getNextID()
	{
		synchronized (tracker)
		{
			return ++mediaTrackerID;
		}
	}

	private int width = -1;
	private int height = -1;

	private URL source;
	private String description;

	private transient Image image;
	private transient int loadStatus = 0;

	public Icon()
	{
		//NOP
	}

	public Icon(URL source)
	{
		load(source);
	}

	public void load(URL source)
	{
		this.source = source;
		image = Toolkit.getDefaultToolkit().getImage(source);
		if (image == null)
		{
			return;
		}
		Object o = image.getProperty("comment", component);
		if (o instanceof String)
		{
			description = (String)o;
		}
		loadImage(image);
	}

	private void loadImage(Image image)
	{
		synchronized (tracker)
		{
			int id = getNextID();

			tracker.addImage(image, id);
			try
			{
				tracker.waitForID(id, 0);
			}
			catch (InterruptedException e)
			{
				System.err.println("Icon: interrupted while loading image (source=" + source + ")");
			}
			loadStatus = tracker.statusID(id, false);
			tracker.removeImage(image, id);

			width = image.getWidth(component);
			height = image.getHeight(component);
		}
	}

	public synchronized void paint(Component c, Graphics g, int x, int y)
	{
		g.drawImage(image, x, y, c);
	}

	/**
	 * Returns the status of the image loading operation.
	 * @return the loading status as defined by java.awt.MediaTracker
	 * @see java.awt.MediaTracker#ABORTED
	 * @see java.awt.MediaTracker#ERRORED
	 * @see java.awt.MediaTracker#COMPLETE
	 */
	public int getImageLoadStatus()
	{
		return loadStatus;
	}

	/**
	 * Returns this icon's <code>Image</code>.
	 * @return the <code>Image</code> object for this icon.
	 */
	public Image getImage()
	{
		return image;
	}

	/**
	 * Sets the image displayed by this icon.
	 * @param image the image
	 */
	public void setImage(Image image)
	{
		this.image = image;
		loadImage(image);
	}

	public String getDescription()
	{
		return description;
	}

	public int getHeight()
	{
		return height;
	}

	public URL getSource()
	{
		return source;
	}

	public int getWidth()
	{
		return width;
	}
}
