/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilegui;

import java.awt.BorderLayout;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.TextEvent;
import java.awt.event.TextListener;
import java.util.Collections;
import java.util.List;
import java.util.Vector;

/**
 * A component to search an ordered list. A query string can be entered in a textfield
 * to find a list item.
 * 
 * @author Michael Hedenus
 */
public class ListChooser extends Panel implements TextListener, ActionListener
{
	private static final long serialVersionUID = 5695249383384640418L;

	private final Vector actionListeners = new Vector();
	private final ListSelection listSelection;
	private final TextField textField;
	private final StringSearcher stringSearcher;

	public ListChooser()
	{
		super(new BorderLayout());
		listSelection = new ListSelection()
		{
			private static final long serialVersionUID = 1L;

			/**
			 * Use the string search to render a list item.
			 */
			protected String toString(List list, Object item, int row, boolean selected)
			{
				return ListChooser.this.stringSearcher.toString(item);
			}
		};
		listSelection.addActionListener(this);

		textField = new TextField();
		textField.addTextListener(this);

		//enter will fire action listener on list selection
		//textField.addActionListener(this);
		textField.addKeyListener(listSelection);

		this.add(listSelection, BorderLayout.CENTER);
		this.add(textField, BorderLayout.NORTH);

		stringSearcher = createStringSearcher();
	}

	public List getList()
	{
		return listSelection.getList();
	}

	/**
	 * Sets the list. The list must be ordered using the comparator returned by 
	 * createStringSearcher() resp. getStringSearcher().
	 */
	public void setList(List list)
	{
		listSelection.setList(list);
	}

	public void addActionListener(ActionListener l)
	{
		actionListeners.add(l);
	}

	public void removeActionListener(ActionListener l)
	{
		actionListeners.remove(l);
	}

	protected void fireActionPerformed()
	{
		if (actionListeners.size() > 0)
		{
			Object item = this.getSelectedItem();
			String command = String.valueOf(item);
			ActionEvent e = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, command);

			for (int i = actionListeners.size() - 1; i >= 0; i--)
			{
				((ActionListener)actionListeners.get(i)).actionPerformed(e);
			}
		}
	}

	/**
	 * Creates the string searcher which is used to display the list and to find an item.
	 */
	public StringSearcher createStringSearcher()
	{
		return new StringSearcher();
	}

	public void textValueChanged(TextEvent e)
	{
		List list = getList();
		if (list != null)
		{
			int i = Collections.binarySearch(list, textField.getText(), stringSearcher);
			if (i < 0)
			{
				//  i = -(insertion point) - 1
				i = -(i + 1);
			}

			if (i < 0)
			{
				i = 0;
			}
			else if (i >= getList().size())
			{
				i = getList().size() - 1;
			}

			listSelection.setSelectionPosition(i);
		}
	}

	public void actionPerformed(ActionEvent e)
	{
		// dispatch action event from ListSelection or Textfield
		fireActionPerformed();
	}

	public Object getSelectedItem()
	{
		return listSelection.getSelectedItem();
	}

	public StringSearcher getStringSearcher()
	{
		return stringSearcher;
	}

	/*
	public boolean isFocusTraversable()
	{
		return true;
	}
	*/
}
