/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilegui;

import java.awt.Canvas;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.ItemSelectable;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

/**
 * A component which renders a list.
 * 
 * @author Michael Hedenus
 */
public class ListSelection extends Canvas implements ItemSelectable, KeyListener, FocusListener
{
	private static final long serialVersionUID = -4171457413204889702L;

	private final Vector itemListeners = new Vector();
	private final Vector actionListeners = new Vector();

	private List list;
	private int scrollPosition = 0;
	private int selectionPosition = 0;
	private Color selectionForeground;
	private Color selectionBackground;
	private Font selectionFont;
	private int cellBorder = 1;
	private Dimension preferredSize = new Dimension(100, 200);

	public ListSelection()
	{
		this.addKeyListener(this);
		this.addFocusListener(this);

		setBackground(UILayout.getInstance().getColor("default.background"));
		setForeground(UILayout.getInstance().getColor("default.foreground"));
		setFont(UILayout.getInstance().getFont("default.font"));

		setSelectionBackground(UILayout.getInstance().getColor("ListSelection.selectionBackground"));
		setSelectionForeground(UILayout.getInstance().getColor("ListSelection.selectionForeground"));
		setSelectionFont(UILayout.getInstance().getFont("ListSelection.selectionFont"));
	}

	public boolean isFocusTraversable()
	{
		return true;
	}

	//temporary variables
	private Rectangle cell = new Rectangle();
	private int visibleRows = 0;

	public void update(Graphics g)
	{
		paint(g);
	}
	
	public void paint(Graphics g)
	{
		int width = getWidth();
		int height = getHeight();
		g.setColor(getBackground());
		g.fillRect(0, 0, width, height);

		if (list == null)
		{
			return;
		}

		FontMetrics fontMetrics = g.getFontMetrics();
		int descent = fontMetrics.getDescent();
		cell.x = 0;
		cell.y = 0;
		cell.height = fontMetrics.getHeight() + 2 * cellBorder;
		cell.width = width;
		visibleRows = 0;

		int maxy = height - (2 * cell.height / 3);
		for (int row = scrollPosition; cell.y < maxy && row < list.size(); row++)
		{
			boolean selected = (row == selectionPosition);
			String line = toString(list, list.get(row), row, selected);
			if (selected)
			{
				g.setColor(selectionBackground);
				g.fillRect(cell.x, cell.y, cell.width - 1, cell.height);
			}

			g.setColor(selected ? selectionForeground : getForeground());
			g.setFont(selected ? selectionFont : getFont());
			g.drawString(line, cell.x + cellBorder, cell.y + cell.height - cellBorder - descent);

			visibleRows++;
			cell.y += cell.height;
		}

		UILayout.getInstance().paintFocusBorder(this, g);
	}

	/**
	 * This method is used to get a string representation of a list item.
	 * Override this method to customize the rendering of an item.
	 * @param list
	 * @param item
	 * @param row
	 * @param selected
	 * @return the string representation of the list item
	 */
	protected String toString(List list, Object item, int row, boolean selected)
	{
		return String.valueOf(item);
	}

	public void setList(List list)
	{
		this.list = list;
	}

	/**
	 * Returns the current number of visible rows (since last paint event).
	 */
	public int getVisibleRows()
	{
		return visibleRows;
	}

	public List getList()
	{
		return list;
	}

	public int getCellBorder()
	{
		return cellBorder;
	}

	public void setCellBorder(int cellBorder)
	{
		this.cellBorder = cellBorder;
	}

	public int getScrollPosition()
	{
		return scrollPosition;
	}

	public int getRows()
	{
		return (list == null ? 0 : list.size());
	}

	public int getSelectionPosition()
	{
		return selectionPosition;
	}

	public void setSelectionPosition(int selectionPosition)
	{
		int oldSelectionPosition = this.selectionPosition;
		if (oldSelectionPosition != selectionPosition)
		{
			this.selectionPosition = selectionPosition;
			scrollRowToVisible(selectionPosition);

			//
			// fire item event
			//
			if (list != null && !itemListeners.isEmpty())
			{
				// deselected
				Object previous = (oldSelectionPosition >= 0 && oldSelectionPosition < list.size() ? list.get(oldSelectionPosition) : null);
				if (previous != null)
				{
					ItemEvent e = new ItemEvent(this, ItemEvent.ITEM_STATE_CHANGED, previous, ItemEvent.DESELECTED);
					for (int i = itemListeners.size() - 1; i >= 0; i--)
					{
						ItemListener l = (ItemListener)itemListeners.get(i);
						l.itemStateChanged(e);
					}
				}

				// selected
				Object current = (selectionPosition >= 0 && selectionPosition < list.size() ? list.get(selectionPosition) : null);
				if (current != null)
				{
					ItemEvent e = new ItemEvent(this, ItemEvent.ITEM_STATE_CHANGED, current, ItemEvent.SELECTED);
					for (int i = itemListeners.size() - 1; i >= 0; i--)
					{
						ItemListener l = (ItemListener)itemListeners.get(i);
						l.itemStateChanged(e);
					}
				}
			}
		}
	}

	private void setScrollPosition(int scrollPosition)
	{
		int rows = getRows();
		if (scrollPosition < 0)
		{
			scrollPosition = 0;
		}
		else if (scrollPosition >= rows)
		{
			scrollPosition = rows - 1;
		}

		int oldScrollPosition = this.scrollPosition;
		if (oldScrollPosition != scrollPosition)
		{
			this.scrollPosition = scrollPosition;
			repaint();
		}
	}

	protected void scrollRowToVisible(int row)
	{
		if (row >= scrollPosition + visibleRows)
		{
			setScrollPosition(row - visibleRows + 1);
		}
		else if (row < scrollPosition)
		{
			setScrollPosition(row);
		}
		else
		{
			repaint();
		}
	}

	public Color getSelectionBackground()
	{
		return selectionBackground;
	}

	public void setSelectionBackground(Color selectionBackground)
	{
		this.selectionBackground = selectionBackground;
	}

	public Color getSelectionForeground()
	{
		return selectionForeground;
	}

	public void setSelectionForeground(Color selectionForeground)
	{
		this.selectionForeground = selectionForeground;
	}

	public Font getSelectionFont()
	{
		return selectionFont;
	}

	public void setSelectionFont(Font selectionFont)
	{
		this.selectionFont = selectionFont;
	}

	public void addItemListener(ItemListener l)
	{
		itemListeners.add(l);
	}

	public void removeItemListener(ItemListener l)
	{
		itemListeners.remove(l);
	}

	public Object getSelectedItem()
	{
		if (list != null && selectionPosition >= 0 && selectionPosition < list.size())
		{
			return list.get(selectionPosition);
		}
		else
		{
			return null;
		}
	}

	public boolean setSelectedItem(Object object)
	{
		boolean result = false;
		if (list != null && object != null)
		{
			int p = 0;
			for (Iterator i = list.iterator(); i.hasNext(); p++)
			{
				Object o = i.next();
				if (o == object || object.equals(o))
				{
					setSelectionPosition(p);
					result = true;
					break;
				}
			}
		}
		return result;
	}

	public Object[] getSelectedObjects()
	{
		return new Object[] { getSelectedItem() };
	}

	public void addActionListener(ActionListener l)
	{
		actionListeners.add(l);
	}

	public void removeActionListener(ActionListener l)
	{
		actionListeners.remove(l);
	}

	protected void fireActionPerformed()
	{
		if (actionListeners.size() > 0)
		{
			Object item = this.getSelectedItem();
			String command = this.toString(list, item, selectionPosition, false);
			ActionEvent e = new ActionEvent(this, ActionEvent.ACTION_PERFORMED, command);

			for (int i = actionListeners.size() - 1; i >= 0; i--)
			{
				((ActionListener)actionListeners.get(i)).actionPerformed(e);
			}
		}
	}

	public void keyPressed(KeyEvent e)
	{
		if(!isEnabled())
		{
			return;
		}
		
		int rows = getRows();

		switch (e.getKeyCode())
		{
			case KeyEvent.VK_HOME:
				setSelectionPosition(0);
				break;

			case KeyEvent.VK_END:
				if (list != null)
				{
					setSelectionPosition(list.size() - 1);
				}
				break;

			case KeyEvent.VK_DOWN:
				//case KeyEvent.VK_RIGHT:
				if (selectionPosition < rows - 1)
				{
					setSelectionPosition(selectionPosition + 1);
				}
				break;

			case KeyEvent.VK_UP:
				//case KeyEvent.VK_LEFT:
				if (selectionPosition > 0)
				{
					setSelectionPosition(selectionPosition - 1);
				}
				break;

			case KeyEvent.VK_PAGE_DOWN:
				int newSelectionDown = selectionPosition + (visibleRows - 1);
				if (newSelectionDown > rows)
				{
					newSelectionDown = rows - 1;
				}
				setSelectionPosition(newSelectionDown);
				break;

			case KeyEvent.VK_PAGE_UP:
				int newSelectionUp = selectionPosition - (visibleRows - 1);
				if (newSelectionUp < 0)
				{
					newSelectionUp = 0;
				}
				setSelectionPosition(newSelectionUp);
				break;

			case KeyEvent.VK_ENTER:
				fireActionPerformed();
				break;

			default:
				break;
		}
	}

	public void keyReleased(KeyEvent e)
	{
		//NOP
	}

	public void keyTyped(KeyEvent e)
	{
		//NOP
	}

	public Dimension getPreferredSize()
	{
		if (preferredSize == null)
		{
			return super.getPreferredSize();
		}
		else
		{
			return preferredSize;
		}
	}

	public void setPreferredSize(Dimension preferredSize)
	{
		this.preferredSize = preferredSize;
	}

	public void focusGained(FocusEvent e)
	{
		repaint();
	}

	public void focusLost(FocusEvent e)
	{
		repaint();
	}
}
