/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilegui;

import java.awt.Canvas;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.ScrollPane;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

/**
 * A canvas which displays a text and does line wrapping.
 * 
 * @author Michael Hedenus
 */
public class TextPanel extends Canvas
{
	private static final long serialVersionUID = -823948466402330521L;

	private class Row
	{
		public final List tokens = new ArrayList();
		public int width;
		public int height;
		public int start;
		public int end;

		public String getToken(int i)
		{
			return (String)tokens.get(i);
		}

		public StringBuffer copyTo(StringBuffer b)
		{
			for (int i = 0; i < tokens.size(); i++)
			{
				b.append(getToken(i));
			}
			return b;
		}
	}

	private class Document
	{
		public final String text;

		public final List rows = new ArrayList();

		public Document(String text)
		{
			this.text = text;
		}

		public Row getRow(int i)
		{
			return (Row)rows.get(i);
		}

		public int getRowCount()
		{
			return rows.size();
		}

		public Dimension updateLayout(FontMetrics fontMetrics, int width)
		{
			Dimension size = new Dimension();
			size.height = 0;
			rows.clear();
			final int rowHeight = fontMetrics.getHeight();
			final String space = " ";
			final int spaceWidth = fontMetrics.stringWidth(space);

			StringTokenizer stringTokenizer = new StringTokenizer(text);
			if (stringTokenizer.hasMoreTokens())
			{
				String token = stringTokenizer.nextToken();

				Row row = new Row();
				row.tokens.add(token);
				row.width = fontMetrics.stringWidth(token);
				row.height = rowHeight;
				this.rows.add(row);
				size.height = row.height + lineGap;

				while (stringTokenizer.hasMoreTokens())
				{
					token = stringTokenizer.nextToken();
					int w = fontMetrics.stringWidth(token);

					if (row.width + w > width)
					{
						row = new Row();
						this.rows.add(row);
						row.height = rowHeight;
						size.height += (row.height + lineGap);
					}
					else
					{
						row.tokens.add(space);
						row.width += spaceWidth;
					}

					row.tokens.add(token);
					row.width += w;

					if (row.width > size.width)
					{
						size.width = row.width;
					}
				}
			}

			return size;
		}

		public String toString()
		{
			StringBuffer b = new StringBuffer();
			for (int i = 0; i < getRowCount(); i++)
			{
				getRow(i).copyTo(b).append('\n');
			}
			return b.toString();
		}
	}

	private Document document;
	private Dimension preferredSize = new Dimension(200, 400);
	private int lineGap = 2;

	public TextPanel()
	{
		setBackground(UILayout.getInstance().getColor("TextPanel.background"));
		setForeground(UILayout.getInstance().getColor("TextPanel.foreground"));
		setFont(UILayout.getInstance().getFont("default.font"));
	}

	public int getLineGap()
	{
		return lineGap;
	}

	public void setLineGap(int lineGap)
	{
		this.lineGap = lineGap;
	}
	
	public int getLineHeight()
	{
		return getFontMetrics(getFont()).getHeight() + getLineGap();
	}

	public String getText()
	{
		if (document != null)
		{
			return document.text;
		}
		else
		{
			return null;
		}
	}

	public void setText(String text)
	{
		document = new Document(text);
		
		int width = Math.max(getWidth(), 0);
		Container parent = getParent();
		if (parent instanceof ScrollPane)
		{
			width = ((ScrollPane)parent).getViewportSize().width;
		}

		// compute the actual size from the given with
		preferredSize = document.updateLayout(getFontMetrics(getFont()), width);
		invalidate();
	}

	public Dimension getPreferredSize()
	{
		return preferredSize;
	}

	public boolean isFocusTraversable()
	{
		return true;
	}
	
	private Rectangle clipRect = new Rectangle();
	private StringBuffer stringBuffer = new StringBuffer();

	public void paint(Graphics g)
	{
		if (document != null)
		{
			//			int width = getWidth();
			int height = getHeight();

			g.setColor(getForeground());
			clipRect = g.getClipBounds(clipRect);

			int y1 = 0;
			int y2 = height;
			Container parent = getParent();
			if (parent instanceof ScrollPane)
			{
				ScrollPane scrollPane = (ScrollPane)parent;
				y1 = scrollPane.getScrollPosition().y;
				y2 = y1 + scrollPane.getViewportSize().height;
			}

			final int rowCount = document.getRowCount();
			int y = 0;
			for (int i = 0; i < rowCount; i++)
			{
				Row row = document.getRow(i);

				if ((y >= y1 - row.height - lineGap) && y <= y2)
				{
					stringBuffer.setLength(0);
					row.copyTo(stringBuffer);
					g.drawString(stringBuffer.toString(), 0, y + row.height);
				}

				y = y + row.height + lineGap;
			}
		}
	}
}
