/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilegui;

import java.awt.EventQueue;
import java.util.Vector;

/**
 * A worker thread which performs jobs for the GUI.
 * 
 * @author Michael Hedenus
 */
public class Worker implements Runnable
{
	private boolean finish = false;
	private final Object semaphore = new Object();
	private final Vector jobs = new Vector();
	private Thread workerThread;

	public Worker()
	{
		//NOP
	}

	public void run()
	{
		// the executinh thread
		workerThread = Thread.currentThread();

		while (!finish)
		{
			// wait for new job
			while (jobs.isEmpty() && !finish)
			{
				synchronized (semaphore())
				{
					try
					{
						System.out.println("Worker: waiting");
						semaphore().wait();
					}
					catch (InterruptedException e)
					{
						//ignore;
					}
				}
			}

			if (!finish)
			{
				// get next job
				Runnable job = null;
				synchronized (semaphore())
				{
					if (!jobs.isEmpty())
					{
						job = (Runnable)jobs.remove(0);
					}
				}

				// do job
				if (job != null)
				{
					execute(job);
				}
			}
		}

		//System.out.println("Worker: finished");
	}

	private void execute(final Runnable job)
	{
		boolean success = false;
		try
		{
			// run the job
			job.run();
			success = true;
		}
		catch (Throwable t)
		{
			handleException(t);
		}
		
		//run done() on event dispatcher
		if (job instanceof Job)
		{
			EventQueue.invokeLater(new Runnable()
			{
				public void run()
				{
					try
					{
						((Job)job).done();
					}
					catch (Throwable t)
					{
						handleException(t);
					}
				}
			});
		}
	}

	/**
	 * Fall-back method to handle an uncaught exception thrown by a job's <code>run()</code> or <code>done()</code> method.
	 */
	public synchronized void handleException(Throwable t)
	{
		t.printStackTrace();
	}

	public Object semaphore()
	{
		return semaphore;
	}

	public void addJob(Runnable job)
	{
		synchronized (semaphore())
		{
			jobs.add(job);
			semaphore().notifyAll();
		}
	}

	public boolean isFinish()
	{
		return finish;
	}

	/**
	 * This will finish the worker and abandon any pending job.
	 * but the currently executed job will be finished. 
	 */
	public synchronized void finish()
	{
		this.finish = true;
		if (workerThread != null)
		{
			workerThread.interrupt();
		}
	}
}
