/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilerss;

import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.Dimension;
import java.awt.EventQueue;
import java.awt.Font;
import java.awt.Frame;
import java.awt.Label;
import java.awt.Panel;
import java.awt.ScrollPane;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.EventObject;
import java.util.List;

import de.hedenus.mobilegui.Animation;
import de.hedenus.mobilegui.Job;
import de.hedenus.mobilegui.ListSelection;
import de.hedenus.mobilegui.Preferences;
import de.hedenus.mobilegui.Tab;
import de.hedenus.mobilegui.Tabs;
import de.hedenus.mobilegui.TextPanel;
import de.hedenus.mobilegui.UILayout;
import de.hedenus.mobilegui.Worker;
import de.hedenus.rdf.Statement;
import de.hedenus.rdfschema.RDFDatabinding;
import de.hedenus.www.Namespaces;

/**
 * The RSS client application.
 * 
 * @author Michael Hedenus
 */
public class RSSClient extends Frame implements Runnable
{
	private static final long serialVersionUID = -6285830851530567439L;

	static final boolean TEST = System.getProperty("test", "false").equalsIgnoreCase("true");

	private RDFDatabinding dataBinding;
	private Preferences preferences = new Preferences("rssclient.properties");
	private Worker worker = new Worker();

	private class ChannelTab extends Tab implements ActionListener
	{
		private static final long serialVersionUID = 477199315166617223L;

		private Animation animation = new Animation();
		private TextField urlField = new TextField();

		public ChannelTab()
		{
			super("Channel", "Channel");

			urlField.addActionListener(this);

			Button openButton = new Button("Open URL");
			openButton.addActionListener(this);

			Panel channelPanel = new Panel(new BorderLayout());
			channelPanel.add(urlField, BorderLayout.CENTER);
			channelPanel.add(openButton, BorderLayout.EAST);

			Panel innerPanel = new Panel(new BorderLayout());
			innerPanel.add(animation, BorderLayout.CENTER);
			innerPanel.add(channelPanel, BorderLayout.NORTH);

			this.add(innerPanel, BorderLayout.CENTER);
		}

		public void tabSelected()
		{
			urlField.requestFocus();
		}

		public void actionPerformed(ActionEvent e)
		{
			try
			{
				URL url = new URL(urlField.getText());
				String channelURI = url.toExternalForm();
				
				openRSS(url, channelURI);
			}
			catch (MalformedURLException ex)
			{
				handleException(ex);
			}
		}
	}

	private class ItemsTab extends Tab
	{
		private static final long serialVersionUID = 5801256130780766188L;

		private ListSelection itemsList = new ListSelection()
		{
			private static final long serialVersionUID = 2439154926808636370L;

			protected String toString(List list, Object o, int row, boolean selected)
			{
				// display the title of the news item
				Item item = (Item)o;
				return item.getTitle();
			}
		};

		public ItemsTab()
		{
			super("Items", "Items");

			itemsList.addActionListener(new ActionListener()
			{
				public void actionPerformed(ActionEvent e)
				{
					Item item = (Item)itemsList.getSelectedItem();
					itemTab.setItem(item);
					tabs.showTab(itemTab);
				}
			});
			this.add(itemsList, BorderLayout.CENTER);
		}

		public void tabSelected()
		{
			itemsList.requestFocus();
		}
	}

	private class ItemTab extends Tab
	{
		private static final long serialVersionUID = 5235413927877627566L;

		private Label title = new Label("");
		private TextPanel description = new TextPanel();

		public ItemTab()
		{
			super("Item", "Item");

			title.setFont(UILayout.getInstance().deriveFromDefaultFont(Font.BOLD));
			//title.setBackground(UILayout.getInstance().getColor("ListSelection.selectionBackground"));
			//title.setForeground(UILayout.getInstance().getColor("ListSelection.selectionForeground"));

			ScrollPane scrollPane = new ScrollPane(ScrollPane.SCROLLBARS_ALWAYS);
			scrollPane.add(description);

			Panel innerPanel = new Panel(new BorderLayout());
			innerPanel.add(title, BorderLayout.NORTH);
			innerPanel.add(scrollPane, BorderLayout.CENTER);

			this.add(innerPanel, BorderLayout.CENTER);
		}

		public void setItem(Item item)
		{
			title.setText(item.getTitle());
			description.setText(item.getDescription());
		}

		public void tabSelected()
		{
			description.requestFocus();
		}
	}

	private Tabs tabs = new Tabs();

	private ChannelTab channelTab = new ChannelTab();
	private ItemsTab itemsTab = new ItemsTab();
	private ItemTab itemTab = new ItemTab();

	public RSSClient()
	{
		super("RSS-Client");
	}

	/**
	 * Executed by the event dispatcher.
	 */
	private void init()
	{
		try
		{
			preferences.open();
		}
		catch (IOException e)
		{
			handleException(e);
		}

		String uri = preferences.getProperty("channel.uri", "http://www.w3.org/2000/08/w3c-synd/home.rss");
		// "http://www.oreillynet.com/pub/feed/20?format=rss1"
		channelTab.urlField.setText(uri);

		tabs.addTab(channelTab);
		tabs.addTab(itemsTab);
		tabs.addTab(itemTab);

		this.add(tabs, BorderLayout.CENTER);
		this.addWindowListener(new java.awt.event.WindowAdapter()
		{
			public void windowClosing(WindowEvent evt)
			{
				exit(evt);
			}
		});
	}

	private synchronized void openRSS(final URL url, final String channelURI)
	{
		channelTab.animation.setAnimated(true);
		tabs.setTabsEnabled(false);

		//
		// load RSS
		//
		worker.addJob(new Job()
		{
			public void run()
			{
				try
				{
					dataBinding = new RDFDatabinding();
					dataBinding.parse(url);
				}
				catch (Exception e)
				{
					worker.handleException(e);
				}
			}

			public void done()
			{
				try
				{
					Channel channel = (Channel)dataBinding.getResource(channelURI);
					if (channel == null)
					{
						// nothing found? query the model for nodes of type 'channel'
						List result = new ArrayList();
						if (dataBinding.getGraph().queryStatement("*", Namespaces.RDF + "type", Channel.URI, result) > 0)
						{
							// take the first channel we find
							String anyChannelURI = ((Statement)result.get(0)).getSubject().about();
							channel = (Channel)dataBinding.getResource(anyChannelURI);
						}
					}

					if (channel != null)
					{
						List items = channel.items();
						itemsTab.itemsList.setList(items);
						preferences.setProperty("channel.uri", channelURI);
					}
				}
				finally
				{
					channelTab.animation.setAnimated(false);
					tabs.setTabsEnabled(true);
					tabs.showTab(itemsTab);
				}
			}
		});
	}

	private void handleException(Throwable t)
	{
		worker.handleException(t);
	}

	/**
	 * Ends the application.
	 */
	private void exit(EventObject event)
	{
		try
		{
			preferences.close();
		}
		catch (IOException e)
		{
			handleException(e);
		}
		finally
		{
			System.exit(0);
		}
	}

	/**
	 * Captures the thread for the worker.
	 */
	private void captureThreadForWorker()
	{
		worker.run();
	}

	/**
	 * Run the application in the event dispatcher
	 */
	public void run()
	{
		init();

		if (TEST)
		{
			// if run in IDE
			setSize(640 - 10, 200 - 40);
			Dimension screen = this.getToolkit().getScreenSize();
			this.setLocation((screen.width - this.getWidth()) / 2, (screen.height - this.getHeight()) / 2);
		}
		else
		{
			this.pack();
		}

		this.setVisible(true);
	}

	/**
	 * Starts the application.
	 */
	public static void main(String args[])
	{
		if (TEST)
		{
			System.out.println("TEST: " + TEST);
		}
		RSSClient rssclient = new RSSClient();
		EventQueue.invokeLater(rssclient);
		rssclient.captureThreadForWorker();
	}
}
