/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilexml;

/**
 * A buffer for characters.
 * 
 * @author Michael Hedenus
 */
public class CharBuffer
{
	private char[] chars;
	private int size;
	private boolean containsOnlyWhitespace = true;

	public CharBuffer(int initialCapacity)
	{
		chars = new char[initialCapacity];
	}

	public CharBuffer append(String s)
	{
		final int len = s.length();
		for (int i = 0; i < len; i++)
		{
			this.append(s.charAt(i));
		}
		
		return this;
	}

	public CharBuffer append(char c)
	{
		chars[size++] = c;
		if (size >= chars.length)
		{
			char[] newChars = new char[chars.length * 2];
			System.arraycopy(chars, 0, newChars, 0, size);
			chars = newChars;
		}
		
		if(containsOnlyWhitespace && !Chars.isWhitespace(c))
		{
			containsOnlyWhitespace = false;
		}
		
		return this;
	}

	public void reset()
	{
		size = 0;
		containsOnlyWhitespace = true;
	}

	public boolean isContainsOnlyWhitespace()
	{
		return containsOnlyWhitespace;
	}

	public int size()
	{
		return size;
	}

	public char charAt(int i)
	{
		return chars[i];
	}

	public String toString()
	{
		return new String(chars, 0, size);
	}

	int getStringCode()
	{
		int code = chars[0];
		for (int i = 1; i < size; i++)
		{
			code = (code << 8) | (chars[i] & 0xFF);
		}
		return code;
	}
	
	/**
	 * Parses the content of the buffer as int.
	 */
	public int parseInt(int radix) throws NumberFormatException
	{
		int result = 0;
		boolean negative = false;
		int i = 0, max = size;
		int limit;
		int multmin;
		int digit;

		if (max > 0)
		{
			if (charAt(0) == '-')
			{
				negative = true;
				limit = Integer.MIN_VALUE;
				i++;
			}
			else
			{
				limit = -Integer.MAX_VALUE;
			}
			multmin = limit / radix;
			if (i < max)
			{
				digit = Character.digit(charAt(i++), radix);
				if (digit < 0)
				{
					throw new NumberFormatException();
				}
				else
				{
					result = -digit;
				}
			}
			while (i < max)
			{
				// Accumulating negatively avoids surprises near MAX_VALUE
				digit = Character.digit(charAt(i++), radix);
				if (digit < 0)
				{
					throw new NumberFormatException();
				}
				if (result < multmin)
				{
					throw new NumberFormatException();
				}
				result *= radix;
				if (result < limit + digit)
				{
					throw new NumberFormatException();
				}
				result -= digit;
			}
		}
		else
		{
			throw new NumberFormatException();
		}
		if (negative)
		{
			if (i > 1)
			{
				return result;
			}
			else
			{ /* Only got "-" */
				throw new NumberFormatException();
			}
		}
		else
		{
			return -result;
		}
	}
}
