/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilexml;

import java.io.IOException;
import java.io.InputStream;

/**
 * A buffered character encoder.
 * 
 * @author Michael Hedenus
 */
public class CharReader
{
	private InputStream inputStream;

	/**
	 * The specified encoding.
	 */
	private String encoding;

	/**
	 * The code of the encoding if it is supported.
	 */
	private int encodingCode;

	/**
	 * The current line number.
	 */
	private int lineNumber;

	/**
	 * UTF-16 endianess.
	 */
	private int endianness;

	private static final int ENDIAN_BIG = 1;
	private static final int ENDIAN_LITTLE = 2;

	public CharReader(int initialStreamBufferSize)
	{
		streamBuffer = new byte[initialStreamBufferSize];
	}

	/**
	 * Sets the stream. If the encoding is null the reader will try to autodetect the encoding. 
	 * @param inputStream the input stream
	 * @param encoding the encoding or null of the encoding is unknown.
	 * @throws IOException
	 */
	public void setInputStream(InputStream inputStream, String encoding) throws IOException
	{
		this.inputStream = inputStream;
		this.endianness = ENDIAN_BIG; /* default */
		setEncoding(encoding);
		resetStreamBuffer();
		fillCharBuffer();
		lineNumber = 1;
	}

	public InputStream getInputStream()
	{
		return inputStream;
	}

	/**
	 * Returns the currently set encoding.
	 * To know which encoding us actually used, see <code>getUsedEncoding</code>.
	 */
	public String getEncoding()
	{
		return encoding;
	}

	/**
	 * Returns the actually used encoding.
	 */
	public String getUsedEncoding()
	{
		return SUPPORTED_ENCODINGS[encodingCode];
	}

	/**
	 * Sets the character encoding. Any unsupported encoding will be treated as <code>UNKNOWN-8BIT</code>.
	 * If the value is null, the reader tries to auto-detect the encoding.
	 * @param encoding the encoding
	 */
	public void setEncoding(String encoding)
	{
		this.encodingCode = 0;
		if (encoding == null)
		{
			this.encoding = null;
		}
		else
		{
			this.encoding = encoding;
			for (int i = 0; i < SUPPORTED_ENCODINGS.length; i++)
			{
				if (SUPPORTED_ENCODINGS[i].equalsIgnoreCase(encoding))
				{
					encodingCode = i;
				}
			}
		}

		// switch UTF-16 to BE or LE
		if (encodingCode == 6)
		{
			switch (endianness)
			{
				case ENDIAN_BIG:
					encodingCode = 4;
					break;

				case ENDIAN_LITTLE:
					encodingCode = 5;
					break;
			}
		}

		//System.out.println("CharReader: encoding: " + getEncoding() + " treated as " + getUsedEncoding());
	}

	////////////////////////////////////////
	//  semantic level 
	////////////////////////////////////////

	/**
	 * Moves the reader <i>before</i> the next non-whitespace char.
	 */
	public void skipWhitespace() throws IOException
	{
		while (Chars.isWhitespace(getChar(1)))
		{
			nextChar();
		}
	}

	/**
	 * Skips n chars.
	 */
	public void skipChars(int count) throws IOException
	{
		for (int i = 0; i < count && hasNextChar(); i++)
		{
			nextChar();
		}
	}

	////////////////////////////////////////
	//  char stream level 
	////////////////////////////////////////

	public static final char EOF_CHAR = (char)(-1);
	private final char[] charBuffer = new char[4];

	/**
	 * Initially fills the char buffer.
	 */
	private void fillCharBuffer() throws IOException
	{
		// start filling with 2nd slot in buffer
		for (int i = 1; i < charBuffer.length; i++)
		{
			charBuffer[i] = readChar();
		}
	}

	/**
	 * Reads and returns the next char.
	 */
	public char nextChar() throws IOException
	{
		//
		// consider LF as new line
		// Note: Apple's \r will not be understood
		//
		if (charBuffer[0] == '\n')
		{
			lineNumber++;
		}
		
		// move the char buffer by one char
		charBuffer[0] = charBuffer[1];
		charBuffer[1] = charBuffer[2];
		charBuffer[2] = charBuffer[3];
		charBuffer[3] = readChar();

		return charBuffer[0];
	}

	/**
	 * Returns true if there are more chars to read.
	 */
	public boolean hasNextChar()
	{
		return charBuffer[1] != EOF_CHAR;
	}

	/**
	 * Returns the i<sup>th</sup> char from the buffer.
	 * @return the i<sup>th</sup> char from the buffer or EOF_CHAR
	 */
	public char getChar(int i)
	{
		return charBuffer[i];
	}

	////////////////////////////////////////
	//  byte stream level 
	////////////////////////////////////////

	public static final String ENCODING_UNKNOWN = "UNKNOWN-8BIT"; /* 0 */
	public static final String ENCODING_ASCII = "US-ASCII"; /* 1 */
	public static final String ENCODING_UTF8 = "UTF-8"; /* 2 */
	public static final String ENCODING_UTF8Y = "UTF-8Y"; /* 3 */
	public static final String ENCODING_UTF16_BE = "UTF-16BE"; /* 4 */
	public static final String ENCODING_UTF16_LE = "UTF-16LE"; /* 5 */
	public static final String ENCODING_UTF16 = "UTF-16"; /* 6 */

	private static final String[] SUPPORTED_ENCODINGS = new String[] { ENCODING_UNKNOWN, ENCODING_ASCII, ENCODING_UTF8, ENCODING_UTF8Y, ENCODING_UTF16_BE, ENCODING_UTF16_LE, ENCODING_UTF16 };

	public int getLineNumber()
	{
		return lineNumber;
	}

	/**
	 * Returns the next char from the stream.
	 * @return the char or EOF_CHAR
	 */
	private char readChar() throws IOException
	{
		char c = EOF_CHAR;

		int byte1 = readByte();
		if (byte1 >= 0)
		{
			switch (encodingCode)
			{
				case 0 /* ENCODING_UNKNOWN */:
				case 1 /* ENCODING_ASCII */:
					c = (char)(byte1 & 0x7F);
					break;

				case 2 /* ENCODING_UTF8 */:
				case 3 /* ENCODING_UTF8Y */:
					c = decodeUTF8(byte1);
					break;

				case 4 /* ENCODING_UTF16_BE */:
					c = decodeUTF16BE(byte1);
					break;

				case 5 /* ENCODING_UTF16_LE */:
					c = decodeUTF16LE(byte1);
					break;
					
				default:
					throw new IOException("unsupported encoding code: "+encodingCode);
			}
		}

		return c;
	}

	private char decodeUTF8(int val) throws IOException
	{
		// count the highest 1 unit 0
		// e.g. 110xxxxx -> 2
		int leading1 = 0;
		for (int mask = 0x80; (val & mask) == mask; mask >>= 1)
		{
			leading1++;
		}

		// inital code value
		int code = (val & (0x7F >> leading1));

		// append trailing bytes
		int trailingBytes = leading1;
		for (int i = 1; i < trailingBytes; i++)
		{
			val = readByte();

			// trailing byte must start with 10xxxxxx
			if ((val & 0xC0) == 0x80)
			{
				// 6 lower bits contain data, append it to code
				code = ((code << 6) | (val & 0x3F));
			}
			else
			{
				throw new IOException("illegal byte in UTF-8 encoding: " + val);
			}
		}

		return (char)code;
	}

	// currently no support for surrogate pair, in fact UCS-2

	private char decodeUTF16BE(int val) throws IOException
	{
		int high = val;
		int low = readByte();
		return (char)((high << 8) | low);
	}

	private char decodeUTF16LE(int val) throws IOException
	{
		int low = val;
		int high = readByte();
		return (char)((high << 8) | low);
	}

	private final byte[] streamBuffer;
	private boolean eof;
	private int streamPosition;
	private int streamBufferPosition;
	private int bytesInBuffer;
	private boolean autoDetectEncoding = false;

	private void resetStreamBuffer()
	{
		eof = false;
		streamPosition = 0;
		streamBufferPosition = 0;
		bytesInBuffer = 0;
		autoDetectEncoding = true;
	}

	/**
	 * Reads the next byte as int or -1 if EOF reached.
	 */
	private int readByte() throws IOException
	{
		if (streamBufferPosition >= bytesInBuffer)
		{
			if (!eof)
			{
				fillBuffer();
			}
			if (eof)
			{
				return -1;
			}
		}
		byte b = streamBuffer[streamBufferPosition++];
		return ((int)b) & 0xFF;
	}

	private void fillBuffer() throws IOException
	{
		streamBufferPosition = 0;
		bytesInBuffer = inputStream.read(streamBuffer);
		if (bytesInBuffer == (-1))
		{
			eof = true;
		}
		else
		{
			streamPosition += bytesInBuffer;
		}

		//
		// UTF-16 specified by external, read possible BOM
		//
		if (autoDetectEncoding && encodingCode >= 4)
		{
			int b0 = (((int)streamBuffer[0]) & 0xFF);
			int b1 = (((int)streamBuffer[1]) & 0xFF);
			int word = ((b0 << 8) | b1);
			if (word == 0xFEFF)
			{
				streamBufferPosition = 2;
				endianness = ENDIAN_BIG;
				setEncoding(ENCODING_UTF16_BE);
			}
			if (word == 0xFFFE)
			{
				streamBufferPosition = 2;
				endianness = ENDIAN_LITTLE;
				setEncoding(ENCODING_UTF16_LE);
			}
		}
		//
		// auto-detect encoding
		//
		else if (autoDetectEncoding && encoding == null)
		{
			if (bytesInBuffer >= 4)
			{
				int b0 = (((int)streamBuffer[0]) & 0xFF);
				int b1 = (((int)streamBuffer[1]) & 0xFF);
				int b2 = (((int)streamBuffer[2]) & 0xFF);
				int b3 = (((int)streamBuffer[3]) & 0xFF);
				int dword = ((b0 << 24) | (b1 << 16) | (b2 << 8) | b3);

				//
				// test UTF-16 BOM
				//
				if ((dword & 0xFFFF0000) == 0xFEFF0000)
				{	
					streamBufferPosition = 2;
					endianness = ENDIAN_BIG;
					setEncoding(ENCODING_UTF16_BE);
				}
				else if ((dword & 0xFFFF0000) == 0xFFFE0000)
				{
					streamBufferPosition = 2;
					endianness = ENDIAN_LITTLE;
					setEncoding(ENCODING_UTF16_LE);
				}
				else
				{
					switch (dword)
					{
						case 0x3C3F786D:
							// starts with <?xm so an encoding may be specified in prolog
							// use UTF-8 as base
							setEncoding(ENCODING_UTF8);
							break;

						case 0xEFBBBF3C:
							// UTF-8 using BOM = 'UTF-8Y' followed by < (any XML markup)
							// this case is deprecated by Unicode spec
							setEncoding(ENCODING_UTF8Y);
							// skip BOM
							streamBufferPosition = 3;
							break;

						case 0x003C003F:
							// UTF-16 BE no BOM
							// this case is strictly speaking not correct
							// see XML spec 4.3.3.
							endianness = ENDIAN_BIG;
							setEncoding(ENCODING_UTF16);
							break;

						case 0x3C003F00:
							// UTF-16 LE no BOM
							// this case is strictly speaking not correct
							// see XML spec 4.3.3.
							endianness = ENDIAN_LITTLE;
							setEncoding(ENCODING_UTF16);
							break;

						default:
							// does not start mit <?xm, MUST be UTF-8
							// or is no XML at all
							setEncoding(ENCODING_UTF8);
							break;
					}
				}
			}
		}

		// this works only of the first time, so reset flag
		autoDetectEncoding = false;
	}
}
