/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilexml;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * An XML parser which does the lexical analysis of a stream.
 * 
 * @author Michael Hedenus
 */
public class XMLParser
{
	private final CharReader reader = new CharReader(256);
	private final CharBuffer charBuffer = new CharBuffer(128);
	private final ArrayList attributeList = new ArrayList();
	private boolean ignoreEmptyText = false;
	private boolean hasProlog = false;
	private boolean finish = false;

	private final XMLParserEventListener eventListener;

	public XMLParser(XMLParserEventListener eventListener)
	{
		this.eventListener = eventListener;
	}

	/**
	 * Parses the input stream.
	 * @param inputStream the input stream
	 * @param encoding the enocding or null if not known (the parse will then try to auto-detect)
	 * @throws IOException
	 */
	public void parse(InputStream inputStream, String encoding) throws IOException
	{
		try
		{
			entities.clear();
			charBuffer.reset();
			entityRefBuffer.reset();
			hasProlog = false;
			finish = false;

			// set stream
			reader.setInputStream(inputStream, encoding);

			//
			// main loop
			//
			while (!finish && reader.hasNextChar())
			{
				parseToken();
			}
		}
		finally
		{
			entities.clear();
		}
	}

	/**
	 * Tells the parser to finish parsing as soon as possible.
	 */
	public void stop()
	{
		finish = true;
	}

	public boolean isIgnoreEmptyText()
	{
		return ignoreEmptyText;
	}

	public void setIgnoreEmptyText(boolean ignoreEmptyText)
	{
		this.ignoreEmptyText = ignoreEmptyText;
	}

	public int getLineNumber()
	{
		return reader.getLineNumber();
	}

	////////////////////
	/// tokens
	////////////////////

	private static final int TEXT = 0;
	private static final int ELEMENT_START = 1;
	private static final int ELEMENT_END = 2;
	private static final int PI = 3;
	private static final int COMMENT_OR_DECLARATION = 4;
	private static final int SECTION = 5;

	private void parseToken() throws IOException
	{
		//look in advance to detect the next token type
		char c1 = reader.getChar(1);
		char c2 = reader.getChar(2);
		char c3 = reader.getChar(3);
		int token = TEXT;

		if (c1 == '<')
		{
			if (c2 == '/')
			{
				// </element>
				token = ELEMENT_END;
			}
			else if (c2 == '?')
			{
				// <? ... ?>
				token = PI;
			}
			else if (c2 == '!')
			{
				if (c3 == '[')
				{
					// <![ CDATA [, INCLUDE [, IGNORE [ ... ]]>
					token = SECTION;
				}
				else
				{
					// <!-- comment -->, <!DOCTYPE, <!ELEMENT, <!ATTLIST, <!ENTITY ...
					token = COMMENT_OR_DECLARATION;
				}
			}
			else
			{
				// <element ... >
				token = ELEMENT_START;
			}
		}

		switch (token)
		{
			case TEXT:
				parseText();
				break;

			case ELEMENT_START:
				parseElementStart();
				break;

			case ELEMENT_END:
				parseElementEnd();
				break;

			case PI:
				parsePI();
				break;

			case COMMENT_OR_DECLARATION:
				parseCommentOrDeclaration();
				break;

			case SECTION:
				parseSection();
				break;
		}
	}

	private void parseText() throws IOException
	{
		String text = parseString('<', ignoreEmptyText);

		if (text != null)
		{
			eventListener.text(text);
		}
	}

	private void parseElementStart() throws IOException
	{
		reader.skipChars(1); /* < */
		String tagName = parseName();

		boolean emptyTag = parseAttributes();

		Element element = new Element(tagName);
		eventListener.elementBegin(element, attributeList);

		// if empty fire element start and immediately element end
		if (emptyTag)
		{
			eventListener.elementEnd(tagName);
		}
	}

	private void parsePI() throws IOException
	{
		reader.skipChars(2); /* <? */
		String pi = parseName();
		parseAttributes();

		if (!hasProlog && "xml".equals(pi))
		{
			hasProlog = true;
			if (attributeList.size() > 0)
			{
				Attribute version = (Attribute)attributeList.get(0);
				if (!"version".equals(version.getName()))
				{
					throw new XMLException("Expected 'version' in prolog.", reader.getLineNumber());
				}

				String declaredEncoding = null;
				if (attributeList.size() > 1)
				{
					Attribute encoding = (Attribute)attributeList.get(1);
					if (!"encoding".equals(encoding.getName()))
					{
						throw new XMLException("Expected 'encoding' in prolog.", reader.getLineNumber());
					}
					else
					{
						declaredEncoding = encoding.getValue();
					}
				}

				if (declaredEncoding != null)
				{
					//TODO: check if the declaredEncoding is compatible with the currently used encoding
					reader.setEncoding(declaredEncoding);
				}
				else
				{
					//if no encoding is declared an no encoding is detected
					//use default UTF-8
					if (reader.getEncoding() == null || reader.getEncoding().equalsIgnoreCase(CharReader.ENCODING_UNKNOWN))
					{
						reader.setEncoding(CharReader.ENCODING_UTF8);
					}
				}
			}
		}
	}

	private String parseName() throws IOException
	{
		charBuffer.reset();
		while (reader.hasNextChar() && Chars.isNameChar(reader.getChar(1)))
		{
			charBuffer.append(reader.nextChar());
		}
		return charBuffer.toString();
	}

	/**
	 * Parses attributes for STag or PI and returns true if the STag is an empty-element tag.
	 */
	private boolean parseAttributes() throws IOException
	{
		attributeList.clear();

		while (reader.hasNextChar())
		{
			reader.skipWhitespace();

			char c1 = reader.getChar(1);
			char c2 = reader.getChar(2);

			if (c1 == '>')
			{
				reader.skipChars(1); /* > */
				return false;
			}
			else if (c1 == '/' && c2 == '>')
			{
				reader.skipChars(2); /* /> */
				return true;
			}
			else if (c1 == '?' && c2 == '>')
			{
				reader.skipChars(2); /* ?> */
				return true;
			}
			else
			{
				// attribute_name S '=' S " value "  
				String attrName = parseName();

				reader.skipWhitespace();
				char eq = reader.nextChar();
				if (eq == '=')
				{
					reader.skipWhitespace();
					char quot = reader.nextChar();

					if (quot == '\"' || quot == '\'')
					{
						String value = parseString(quot, false); /* parse until same quot as begun */
						reader.skipChars(1); /* ending quot */

						Attribute attribute = new Attribute(attrName, value);
						attributeList.add(attribute);
					}
					else
					{
						throw new XMLException("Expected \" or \'.", reader.getLineNumber());
					}
				}
				else
				{
					throw new XMLException("Expected '=' after attribute '" + attrName + "'.", reader.getLineNumber());
				}
			}
		}

		return false;
	}

	private void parseElementEnd() throws IOException
	{
		reader.skipChars(2); /* </ */
		String tagName = parseName();
		reader.skipWhitespace();
		reader.skipChars(1); /* > */

		eventListener.elementEnd(tagName);
	}

	/**
	 * &lt;! detected.
	 */
	private void parseCommentOrDeclaration() throws IOException
	{
		reader.skipChars(2); /* <! */

		char c1 = reader.getChar(1);
		char c2 = reader.getChar(2);
		if (c1 == '-' && c2 == '-')
		{
			parseComment();
		}
		else
		{
			parseDeclaration();
		}
	}

	/**
	 * Parsing &lt;!-- -- &gt;
	 */
	private void parseComment() throws IOException
	{
		reader.skipChars(2); /* -- */

		//System.out.print("#comment: ");

		while (reader.hasNextChar())
		{
			char c0 = reader.nextChar();
			char c1 = reader.getChar(1);
			char c2 = reader.getChar(2);

			if (c0 == '-' && c1 == '-' && c2 == '>')
			{
				reader.skipChars(2); /* ->, first '-' already consumed */
				break;
			}

			//System.out.print(c0);
		}

		//System.out.println();
	}

	/**
	 * Parsing keyword after &lt;!.
	 */
	private void parseDeclaration() throws IOException
	{
		String token = parseName();

		if ("DOCTYPE".equals(token))
		{
			parseDoctype();
		}
		if ("ENTITY".equals(token))
		{
			parseEntityDeclaration();
		}
	}

	/**
	 * Parsing after &lt;!DOCTYPE until ]&gt;.
	 */
	private void parseDoctype() throws IOException
	{
		reader.skipWhitespace();
		String rootElementName = parseName();

		reader.skipWhitespace();
		char c0 = reader.nextChar();
		if (c0 != '[')
		{
			throw new XMLException("Expected '['.", reader.getLineNumber());
		}

		while (!finish && reader.hasNextChar())
		{
			reader.skipWhitespace();

			char c1 = reader.getChar(1);
			char c2 = reader.getChar(2);
			if (c1 == '<' && c2 == '!')
			{
				parseCommentOrDeclaration();
			}
			if (c1 == ']' && c2 == '>')
			{
				// end of DTD
				break;
			}
		}

		reader.skipChars(2); /* ]> */
	}

	/**
	 * Parsing after &lt;!ENTITY until &gt;.
	 */
	private void parseEntityDeclaration() throws IOException
	{
		reader.skipWhitespace();
		String entityName = parseName();

		reader.skipWhitespace();
		char quot = reader.nextChar();
		if (quot != '\"')
		{
			throw new XMLException("Expected '\"'.", reader.getLineNumber());
		}
		String entityValue = parseString('\"', false);
		reader.skipChars(1); /* " */

		reader.skipWhitespace();
		char br = reader.nextChar();
		if (br != '>')
		{
			throw new XMLException("Expected '>'.", reader.getLineNumber());
		}

		entities.put(entityName, entityValue);
	}

	private void parseSection() throws IOException
	{
		reader.skipChars(3); /* <![ */
		String sectionName = parseName();
		if ("CDATA".equals(sectionName))
		{
			if (reader.nextChar() != '[')
			{
				throw new XMLException("Expected '[' after '<![CDATA'.", getLineNumber());
			}

			charBuffer.reset();
			while (reader.hasNextChar())
			{
				char c0 = reader.nextChar();
				char c1 = reader.getChar(1);
				char c2 = reader.getChar(2);

				if (c0 == ']' && c1 == ']' && c2 == '>')
				{
					reader.skipChars(2); /* ->, first '-' already consumed */
					break;
				}
				else
				{
					charBuffer.append(c0);
				}
			}

			// callback
			eventListener.text(charBuffer.toString());

		}
		else
		{
			throw new XMLException("Unsupported section name: '" + sectionName + "'.", getLineNumber());
		}
	}

	/**
	 * Parses text until the terminating char or EOF is reached.
	 * The parse will stop <i>before</i> the terminating char.
	 * Returns null if the charBuffer contains only whitespace and ignore empty text is set to true.
	 */
	private String parseString(char terminatingChar, boolean ignoreEmptyString) throws IOException
	{
		charBuffer.reset();

		while (reader.hasNextChar() && reader.getChar(1) != terminatingChar)
		{
			char c0 = reader.nextChar();

			if (c0 == '&')
			{
				parseEntityRef();
			}
			else
			{
				charBuffer.append(c0);
			}
		}

		if (ignoreEmptyString && charBuffer.isContainsOnlyWhitespace())
		{
			return null;
		}
		else
		{
			return charBuffer.toString();
		}
	}

	//////////////////////////
	/// entities
	//////////////////////////

	private final HashMap entities = new HashMap();

	private static final int GT = ('g' << 8) | 't';
	private static final int LT = ('l' << 8) | 't';
	private static final int AMP = ('a' << 16) | ('m' << 8) | 'p';
	private static final int APOS = ('a' << 24) | ('p' << 16) | ('o' << 8) | 's';
	private static final int QUOT = ('q' << 24) | ('u' << 16) | ('o' << 8) | 't';

	/**
	 * Buffer used to parse the entity reference.
	 */
	private final CharBuffer entityRefBuffer = new CharBuffer(5);

	/**
	 * Translate an entity reference.
	 */
	private void parseEntityRef() throws IOException
	{
		entityRefBuffer.reset();

		//c0 = '&'
		char c1 = reader.getChar(1);
		char c2 = reader.getChar(2);
		if (c1 == '#')
		{
			// char number 
			int radix;
			if (c2 == 'x')
			{
				// hex number
				reader.skipChars(2); /* skip #x */
				parseEntityName();
				radix = 16;
			}
			else
			{
				// decimal number
				reader.skipChars(1); /* skip # */
				parseEntityName();
				radix = 10;
			}

			try
			{
				char unicode = (char)entityRefBuffer.parseInt(radix);
				charBuffer.append(unicode);
			}
			catch (NumberFormatException nfex)
			{
				nfex.printStackTrace();
			}
		}
		else
		{
			// named entity
			parseEntityName();
			int code = 0;
			if (entityRefBuffer.size() >= 2 && entityRefBuffer.size() <= 4)
			{
				code = entityRefBuffer.getStringCode();
			}
			switch (code)
			{
				case GT:
					charBuffer.append('>');
					break;

				case LT:
					charBuffer.append('<');
					break;

				case AMP:
					charBuffer.append('&');
					break;

				case APOS:
					charBuffer.append('\'');
					break;

				case QUOT:
					charBuffer.append('\"');
					break;

				default:
					// try to get named entity
					String entityRef = entityRefBuffer.toString();
					String entity = (String)entities.get(entityRef);
					if (entity != null)
					{
						charBuffer.append(entity);
					}
					else
					{
						throw new XMLException("Cannot resolve entity '" + entityRef + "'.", reader.getLineNumber());
					}
					break;
			}
		}
	}

	/**
	 * Parses the name of the entity <code>&amp;NAME;</code> into the <code>entityRefBuffer</code>.
	 */
	private void parseEntityName() throws IOException
	{
		while (reader.hasNextChar() && reader.nextChar() != ';')
		{
			entityRefBuffer.append(reader.getChar(0));
		}
	}
}
