/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.mobilexml;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import de.hedenus.www.Namespaces;
import de.hedenus.www.URIUtil;

/**
 * An XML reader which understands namespaces, xml:id, xml:base, xml:space, xml:lang.
 * The XMLReader is based on XMLParser.
 * 
 * @author Michael Hedenus
 */
public class XMLReader implements XMLParserEventListener
{
	/**
	 * The scope of an XML element.
	 * 
	 * @author Michael Hedenus
	 */
	public static class Scope
	{
		/**
		 * "preserve".
		 */
		public static final String SPACE_PRESERVE = "preserve";
		
		/**
		 * "default".
		 */
		public static final String SPACE_DEFAULT = "default";

		/**
		 * The element is empty.
		 */
		public static final int CONTENT_EMPTY = 0;

		/**
		 * The element contains only elements (or only whitespace if 'ignoreEmptyText' is set to true).
		 */
		public static final int CONTENT_ELEMENTS = 1;

		/**
		 * The element contains only text and no child elements.
		 */
		public static final int CONTENT_TEXT = 2;

		/**
		 * The element contains 'mixed content' i.e. text and child elements.
		 */
		public static final int CONTENT_MIXED = CONTENT_ELEMENTS | CONTENT_TEXT;

		private final Scope parent;
		private final int depth;
		private String lang;
		private String base;
		private String space;
		private String id;
		private String defaultNamespace;
		private final HashMap namespaceMap = new HashMap();
		private Element element;
		private Attribute[] attributes;
		private int content = CONTENT_EMPTY;

		protected Scope(Scope parent)
		{
			this.parent = parent;
			this.depth = (parent == null ? 0 : parent.depth + 1);
		}

		protected void reset()
		{
			lang = null;
			base = null;
			space = null;
			id = null;
			defaultNamespace = null;
			namespaceMap.clear();
			element = null;
			attributes = null;
			content = CONTENT_EMPTY;
		}

		public String toString()
		{
			return element + ": defaultNS=" + String.valueOf(getDefaultNamespace()) + //
					", base=" + String.valueOf(getBase()) + //
					", lang=" + String.valueOf(getLang()) + //
					", space=" + String.valueOf(getSpace()); //
		}

		public Scope getParent()
		{
			return parent;
		}

		public int getDepth()
		{
			return depth;
		}

		public Element getElement()
		{
			return element;
		}

		public Attribute[] getAttributes()
		{
			return attributes;
		}

		/**
		 * Resolves an URI against the current base.
		 * @param relURI
		 * @return the resolved URI
		 */
		public String resolveURI(String relURI)
		{
			return URIUtil.resolve(relURI, getBase());
		}

		public String getBase()
		{
			if (this.base == null && this.parent != null)
			{
				return parent.getBase();
			}
			return this.base;
		}

		public String getLang()
		{
			if (this.lang == null && this.parent != null)
			{
				return parent.getLang();
			}
			return this.lang;
		}

		public String getSpace()
		{
			if (this.space == null && this.parent != null)
			{
				return parent.getSpace();
			}
			return this.space;
		}

		public String getId()
		{
			return id;
		}

		public String getDefaultNamespace()
		{
			if (this.defaultNamespace == null && this.parent != null)
			{
				return parent.getDefaultNamespace();
			}
			return this.defaultNamespace;
		}
		
		public HashMap getNamespaceMap()
		{
			return namespaceMap;
		}

		public String getNamespace(String prefix)
		{
			if (prefix.equals(""))
			{
				return getDefaultNamespace();
			}

			if (prefix.equals("xml"))
			{
				return Namespaces.XMLNS;
			}

			if (prefix.equals("xmlns"))
			{
				return Namespaces.XMLNS_ATTRIBUTE;
			}

			return getNamespaceInherited(prefix);
		}

		private String getNamespaceInherited(String prefix)
		{
			String namespace = (String)this.namespaceMap.get(prefix);
			if (namespace == null && this.parent != null)
			{
				return parent.getNamespaceInherited(prefix);
			}

			return namespace;
		}

		private void defineNamespace(String prefix, String namespaceURI)
		{
			if (prefix.equals(""))
			{
				defaultNamespace = namespaceURI;
			}
			else
			{
				//System.out.println(prefix+"="+namespaceURI);
				namespaceMap.put(prefix, namespaceURI);
			}
		}

		/**
		 * The content of the element. This property is only available at <code>elementEnd()</code>.
		 */
		public int getContent()
		{
			return content;
		}
	}

	private class ScopeStack
	{
		private ArrayList scopes = new ArrayList();
		private int sp;

		/**
		 * Returns the new scope.
		 */
		public Scope push()
		{
			// increase count
			sp++;

			// if count is larger than scope cache
			// we have to create a new instance
			if (sp > scopes.size())
			{
				Scope parent = null;
				if (sp > 1)
				{
					// if we have 2 or more scopes
					// we have to link the new scope with its parent
					parent = (Scope)scopes.get(sp - 2);
				}
				scopes.add(createScope(parent));
			}

			return peek();
		}

		public void pop()
		{
			// if there is a scope, it must be cleared
			Scope scope = peek();
			if (scope != null)
			{
				scope.reset();
				sp--;
			}
		}

		public Scope peek()
		{
			if (sp > 0)
			{
				return (Scope)scopes.get(sp - 1);
			}
			else
			{
				return null;
			}
		}

		public void reset()
		{
			sp = 0;
			scopes.clear();
		}
	}

	/**
	 * Creates a new scope instance.
	 * Override this method if you want to create a subclass of the scope.
	 */
	protected Scope createScope(Scope parent)
	{
		return new Scope(parent);
	}

	private final XMLParser parser = new XMLParser(this);
	private final XMLReaderEventListener listener;
	private final ScopeStack stack = new ScopeStack();
	private final ArrayList visibleAttributes = new ArrayList();

	public XMLReader(XMLReaderEventListener listener)
	{
		this.listener = listener;
	}

	/**
	 * Tells the reader to finish parsing as soon as possible.
	 */
	public void stop()
	{
		parser.stop();
	}

	public int getLineNumber()
	{
		return parser.getLineNumber();
	}

	public void reset()
	{
		stack.reset();
	}

	public void read(InputStream in, String encoding, String documentBase) throws IOException, XMLException
	{
		try
		{
			// just in case...
			reset();
			
			// push scope for the document (above root element)
			Scope documentScope = stack.push();
			documentScope.defaultNamespace = Namespaces.NO_NAMESPACE;
			documentScope.base = documentBase;

			// run
			parser.parse(in, encoding);
		}
		finally
		{
			// clean up
			reset();
		}
	}

	public boolean isIgnoreEmptyText()
	{
		return parser.isIgnoreEmptyText();
	}

	public void setIgnoreEmptyText(boolean ignoreEmptyText)
	{
		parser.setIgnoreEmptyText(ignoreEmptyText);
	}

	public void elementBegin(Element element, List attributeList) throws XMLException
	{
		Scope parentScope = stack.peek();
		parentScope.content |= Scope.CONTENT_ELEMENTS;

		Scope scope = stack.push();

		//
		// pre-parse
		//
		visibleAttributes.clear();
		for (int i = 0; i < attributeList.size(); i++)
		{
			final Attribute attribute = (Attribute)attributeList.get(i);
			final String name = attribute.name;
			final String value = attribute.value;

			if (name.startsWith("xml"))
			{
				// namespace declaration
				if (name.startsWith("xmlns"))
				{
					String prefix = "";
					int colon = name.indexOf(':');
					if (colon > 0)
					{
						prefix = name.substring(colon + 1);
					}

					scope.defineNamespace(prefix, value);
				}

				// special attributes
				if ("xml:id".equals(name))
				{
					scope.id = value;
				}
				else if ("xml:base".equals(name))
				{
					// get parent base
					String parentBase = scope.parent.getBase();

					// resolve base against parent base
					scope.base = URIUtil.resolve(value, parentBase);
				}
				else if ("xml:lang".equals(name))
				{
					scope.lang = value;
				}
				else if ("xml:space".equals(name))
				{
					if ("preserve".equals(value))
					{
						scope.space = Scope.SPACE_PRESERVE;
					}
					else if ("default".equals(value))
					{
						scope.space = Scope.SPACE_DEFAULT;
					}
					else
					{
						throw new XMLException("Unsupported xml:space attibute value: '"+value+"'.", getLineNumber());
					}
				}
			}
			else
			{
				visibleAttributes.add(attribute);
			}
		}

		//
		// element
		// 
		scope.element = element;
		int colon = element.name.indexOf(':');
		if (colon > 0)
		{
			element.localName = element.name.substring(colon + 1);
			element.namespacePrefix = element.name.substring(0, colon);
		}
		else
		{
			element.localName = element.name;
			element.namespacePrefix = "";
		}
		element.namespaceURI = scope.getNamespace(element.namespacePrefix);

		//
		// attributes
		//
		scope.attributes = new Attribute[visibleAttributes.size()];
		for (int i = 0; i < visibleAttributes.size(); i++)
		{
			Attribute attribute = (Attribute)visibleAttributes.get(i);

			colon = attribute.name.indexOf(':');
			if (colon > 0)
			{
				attribute.localName = attribute.name.substring(colon + 1);
				attribute.namespacePrefix = attribute.name.substring(0, colon);
				attribute.namespaceURI = scope.getNamespace(attribute.namespacePrefix);
			}
			else
			{
				//if the attribute is not explicitely bound to a namespace it
				// is bound to the element's namespace
				attribute.localName = attribute.name;
				attribute.namespacePrefix = null;
				attribute.namespaceURI = null;
			}

			scope.attributes[i] = attribute;
		}

		//
		// callback
		//
		listener.elementBegin(scope.element, scope);
	}

	public void elementEnd(String tagName) throws XMLException
	{
		Scope scope = stack.peek();

		if (!scope.element.getName().equals(tagName))
		{
			throw new XMLException("Missing end tag for element '" + scope.element.getName() + "'.", getLineNumber());
		}

		listener.elementEnd(scope.element, scope);
		stack.pop();
	}

	public void text(String text) throws XMLException
	{
		listener.text(text, stack.peek());
		Scope scope = stack.peek();
		scope.content |= Scope.CONTENT_TEXT;
	}
}
