/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.owl;

import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;

import de.hedenus.rdf.RDFConstants;
import de.hedenus.rdf.RDFGraph;
import de.hedenus.rdf.Statement;
import de.hedenus.rdfschema.InferenceEngine;
import de.hedenus.rdfschema.RDFDatabinding;
import de.hedenus.rdfschema.RDFSchemaConstants;
import de.hedenus.rdfschema.Rule;
import de.hedenus.rdfschema.rules.PropertyRule;
import de.hedenus.rdfschema.rules.ResourceRule;
import de.hedenus.rdfschema.rules.SubclassOfResourceRule;
import de.hedenus.rdfschema.rules.PropertyPropagationRule;
import de.hedenus.rdfschema.rules.TypePropagationRule;

/**
 * OWL implementation interface.
 * 
 * @author Michael Hedenus
 */
public class OWLImplementation extends RDFDatabinding
{
	private final InferenceEngine inferenceEngine;

	public OWLImplementation()
	{
		getTypeMapping().add(Ontology.URI, Ontology.class.getName());

		inferenceEngine = new InferenceEngine(this);

		try
		{
			initSchemas();
		}
		catch (IOException e)
		{
			throw new OWLException("Cannot import basic schemas: " + e.getMessage());
		}
		
		addRule(new ResourceRule());
		addRule(new PropertyRule());
		addRule(new SubclassOfResourceRule());
		addRule(new TypePropagationRule());
		addRule(new PropertyPropagationRule());
	}

	private void initSchemas() throws IOException
	{
		this.parse(RDFSchemaConstants.SCHEMA_URL, RDFGraph.STATEMENT_AXIOM);
		this.parse(OWLConstants.SCHEMA_URL, RDFGraph.STATEMENT_AXIOM);
	}

	public void addRule(Rule rule)
	{
		inferenceEngine.addRule(rule);
	}

	public InferenceEngine getInferenceEngine()
	{
		return inferenceEngine;
	}

	/**
	 * Imports an OWL ontology.
	 * @param url
	 * @throws IOException
	 */
	public void importOntology(URL url) throws IOException
	{
		parse(url);
	}

	/**
	 * Queries for all imported ontologies.
	 */
	public Ontology[] getOntologies()
	{
		ArrayList results = new ArrayList();
		int r = getGraph().queryStatement("*", RDFConstants.TYPE_URI, Ontology.URI, results);
		Ontology[] ontologies = new Ontology[r];

		for (int i = 0; i < r; i++)
		{
			Statement s = (Statement)results.get(i);
			ontologies[i] = (Ontology)getResource(s.getSubject().about());
		}

		return ontologies;
	}
}
