/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf;

import de.hedenus.graph.Arc;
import de.hedenus.graph.Node;
import java.io.Serializable;

/**
 * A basic (un-reified) statement.
 * 
 * @author Michael Hedenus
 */
public class BasicStatement implements Statement, Serializable, Comparable
{
	private static final long serialVersionUID = 8788222599918182943L;

	private final String id;
	private final RDFNode subject;
	private final String predicate;
	private final RDFNode object;
	private final int hashCode;

	protected BasicStatement(RDFNode subject, String predicate, RDFNode object, String id)
	{
		if (subject == null)
		{
			throw new IllegalArgumentException("subject must not be null");
		}
		if (predicate == null)
		{
			throw new IllegalArgumentException("predicate must not be null");
		}
		if (object == null)
		{
			throw new IllegalArgumentException("object must not be null");
		}
		if (id == null)
		{
			throw new IllegalArgumentException("id must not be null");
		}
		if (!(subject instanceof BlankNode || subject instanceof URIRefNode))
		{
			throw new RDFException("Subject of a statement must be either resource or blank node.");
		}

		this.subject = subject;
		this.predicate = predicate;
		this.object = object;
		this.id = id;

		int hash = getSubject().hashCode();
		hash = 31 * hash + getPredicate().hashCode();
		hash = 31 * hash + getObject().hashCode();
		this.hashCode = hash;
	}

	public RDFNode getSubject()
	{
		return subject;
	}

	public String getPredicate()
	{
		return predicate;
	}

	public RDFNode getObject()
	{
		return object;
	}

	public Node leftNode()
	{
		return getSubject();
	}

	public Node rightNode()
	{
		return getObject();
	}

	public int orientation()
	{
		return Arc.LEFT_TO_RIGHT;
	}

	/**
	 * The statements's label is the predicate URI.
	 */
	public String label()
	{
		return predicate;
	}

	public String id()
	{
		return id;
	}

	public String toString()
	{
		return getSubject().toString() + " <" + predicate + "> " + getObject().toString() + " .";
	}

	/**
	 * Two statements are equal if their subjects, predicates and objects are equal.
	 */
	public boolean equals(Object obj)
	{
		return compareTo(obj) == 0;
	}

	public int hashCode()
	{
		return hashCode;
	}

	public int compareTo(Object obj)
	{
		Statement that = (Statement)obj;
		
		if (this == that)
		{
			return 0;
		}
		int s = this.getSubject().compareTo(that.getSubject());
		if (s == 0)
		{
			int p = this.getPredicate().compareTo(that.getPredicate());
			if (p == 0)
			{
				int o = this.getObject().compareTo(that.getObject());
				return o;
			}
			else
			{
				return p;
			}
		}
		else
		{
			return s;
		}
	}
}
