/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf;

import java.io.Serializable;

/**
 * A RDF <a href="http://www.w3.org/TR/rdf-concepts/#section-Graph-Literal">literal</a>.
 * 
 * @author Michael Hedenus
 */
public class Literal extends RDFNode implements Serializable
{
	private static final long serialVersionUID = 2236194900247342692L;

	private final String string;
	private final String datatype;
	private final String language;
	private final int hashCode;

	protected Literal(String string, String datatype, String language, String id)
	{
		super(id);
		if (string == null)
		{
			throw new IllegalArgumentException("string must not be null");
		}
		this.string = string;
		this.datatype = datatype;
		this.language = language;

		int hash = string.hashCode();
		if (datatype != null)
		{
			hash = 31 * hash + datatype.hashCode();
		}
		if (language != null)
		{
			hash = 31 * hash + language.hashCode();
		}
		this.hashCode = hash;
	}

	/**
	 * Two literals are equal if:
	 * (rdf-concepts 6.5.1 Literal Equality)
	 * <ul>
	 * <li>The strings of the two lexical forms compare equal, character by character.</li>
	 * <li>Either both or neither have language tags.</li>
	 * <li>The language tags, if any, compare equal.</li>
	 * <li>Either both or neither have datatype URIs.</li>
	 * <li>The two datatype URIs, if any, compare equal, character by character.</li>
	 * </ul>
	 */
	public int compareTo(Object obj)
	{
		if (obj instanceof Literal)
		{
			if (this == obj)
			{
				return 0;
			}
			else
			{
				Literal that = (Literal)obj;

				final int s = StringUtil.compare(this.string, that.string);
				if (s == 0)
				{
					final int d = StringUtil.compare(this.datatype, that.datatype);
					if (d == 0)
					{
						final int l = StringUtil.compare(this.language, that.language);
						return l;
					}
					else
					{
						return d;
					}
				}
				else
				{
					return s;
				}
			}
		}
		else
		{
			return super.compareTo(obj);
		}
	}

	public String getString()
	{
		return string;
	}

	public int hashCode()
	{
		return hashCode;
	}

	public String getDatatype()
	{
		return datatype;
	}

	public String getLanguage()
	{
		return language;
	}

	public String about()
	{
		return id();
	}

	public String label()
	{
		return "\"" + string + "\"" + (language != null ? "@" + language : "") + (datatype != null ? ("^^<" + datatype + ">") : "");
	}
}
