/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf;

import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.TreeSet;

import de.hedenus.graph.Graph;
import de.hedenus.www.NamespaceContext;

/**
 *  A RDF graph.
 *  
 *  @author Michael Hedenus
 */
public class RDFGraph implements Graph, Serializable
{
	private static final long serialVersionUID = 2918151624834976710L;

	private int idCounter = 0;

	private final NamespaceContext namespaceContext = new NamespaceContext(true);
	private final HashMap nodes = new HashMap();

	public static final int STATEMENT_AXIOM = 1;
	public static final int STATEMENT_ASSERTED = 2;
	public static final int STATEMENT_INFERED = 4;

	private static class StatementHolder implements Serializable
	{
		private static final long serialVersionUID = 299013562420612358L;

		public final Statement statement;
		public int type;
		public int count = 1;

		public StatementHolder(Statement statement)
		{
			this.statement = statement;
		}

		public String toString()
		{
			return statement.getClass().getName() + "[type=" + type + ", count=" + count + "]";
		}
	}

	private List graphListeners = new ArrayList();

	/**
	 * Maps statement to StatementMetaInfo
	 */
	private final HashMap statements = new HashMap();

	private final SPIndex spIndex = new SPIndex(256);

	public RDFGraph()
	{
		//NOP
	}

	public void addGraphListener(GraphListener l)
	{
		graphListeners.add(l);
	}

	public void removeGraphListener(GraphListener l)
	{
		graphListeners.remove(l);
	}

	/**
	 * Returns a node by its key.
	 * <ol>
	 * <li>Blank Node: id()</li>
	 * <li>Literal: id()</li>
	 * <li>Resource: URI external form</li>
	 * </ol>
	 * @param key the node's key.
	 */
	public RDFNode getNode(String key)
	{
		return (RDFNode)nodes.get(key);
	}

	public Collection nodes()
	{
		return nodes.values();
	}

	public Collection arcs()
	{
		return statements.keySet();
	}

	public NamespaceContext getNamespaceContext()
	{
		return namespaceContext;
	}

	public void dump()
	{
		dump(System.out, true);
	}

	public void dump(OutputStream outputStream, boolean sort)
	{
		PrintWriter pw = new PrintWriter(outputStream);

		Collection statements = sort ? new TreeSet(arcs()) : arcs();
		for (Iterator i = statements.iterator(); i.hasNext();)
		{
			Statement statement = (Statement)i.next();
			pw.println(statement);
		}

		pw.flush();
	}

	private static final String WILDCARD = "*";

	/**
	 * Queries statement and puts the matching statements into <code>result</code>.
	 * @param subject
	 * @param predicate
	 * @param object
	 * @param result
	 * @return the number of matches found
	 */
	public int queryStatement(String subject, String predicate, String object, Collection result)
	{
		int matches = 0;

		//
		// if the query is of the from <subject> <predicate> *, then the SP-Index can be used
		//
		if (!WILDCARD.equals(subject) && !WILDCARD.equals(predicate) && WILDCARD.equals(object))
		{
			Statement[] s = spIndex.get(subject, predicate);
			if (s != null)
			{
				for (int i = 0; i < s.length; i++)
				{
					result.add(s[i]);
				}
				matches = s.length;
			}
		}
		else
		{
			for (Iterator i = arcs().iterator(); i.hasNext();)
			{
				Statement statement = (Statement)i.next();

				if (matches(subject, statement.getSubject()) && matches(predicate, statement.getPredicate()) && matches(object, statement.getObject()))
				{
					result.add(statement);
					matches++;
				}
			}
		}

		return matches;
	}

	private boolean matches(String query, Object value)
	{
		if (WILDCARD.equals(query))
		{
			return true;
		}
		else if (value instanceof URIRefNode)
		{
			return ((URIRefNode)value).getURI().equals(query);
		}
		else if (value instanceof Literal)
		{
			return ((Literal)value).getString().equals(query);
		}
		else if (value instanceof BlankNode)
		{
			return ((BlankNode)value).id().equals(query);
		}
		else if (value instanceof String)
		{
			return ((String)value).equals(query);
		}
		else
		{
			return false;
		}
	}

	///////////////////////////////////////////////////////////////
	//// Factory methods
	///////////////////////////////////////////////////////////////

	/**
	 * Creates a new ID.
	 */
	private String createID()
	{
		return "ID" + (++idCounter);
	}

	public int getStatementType(Statement statement)
	{
		StatementHolder metaInfo = getStatementHolder(statement);
		if(metaInfo!=null)
		{
			return metaInfo.type;
		}
		else
		{
			throw new RDFException("Statement not found: "+String.valueOf(statement));
		}
	}
	
	/**
	 * Creates a (basic) statement.
	 */
	public Statement createStatement(RDFNode subject, String predicate, RDFNode object, int statementType)
	{
		Statement statement = new BasicStatement(subject, predicate, object, createID());

		StatementHolder metaInfo = getStatementHolder(statement);
		if (metaInfo == null)
		{
			// new statement
			metaInfo = new StatementHolder(statement);
			metaInfo.type = statementType;

			statements.put(statement, metaInfo);

			//add to SP-Index
			spIndex.add(statement);

			// callback
			for (int i = 0; i < graphListeners.size(); i++)
			{
				((GraphListener)graphListeners.get(i)).statementCreated(this, statement);
			}
		}
		else
		{
			//return existing statement object
			statement = metaInfo.statement;

			// decrease id counter
			idCounter--;

			//count duplicates
			metaInfo.count++;
		}

		return statement;
	}

	private StatementHolder getStatementHolder(Statement statement)
	{
		return (StatementHolder)statements.get(statement);
	}

	/**
	 * Returns or lazily creates a resource object for a URI.
	 * @return the Resource instance.
	 */
	public URIRefNode createURIRefNode(String uri) throws RDFException
	{
		RDFNode node = getNode(uri);
		if (node == null)
		{
			node = new URIRefNode(uri, createID());
			nodes.put(uri, node);
		}

		if (node instanceof URIRefNode)
		{
			return (URIRefNode)node;
		}
		else
		{
			throw new RDFException("Element is contained by the graph but it is not a resource.");
		}
	}

	/**
	 * Creates a new blank node.
	 */
	public BlankNode createBlankNode()
	{
		BlankNode node = new BlankNode(createID());
		nodes.put(node.id(), node);
		return node;
	}

	/**
	 * Creates a new literal node.
	 */
	public Literal createLiteral(String string, String datatype, String language)
	{
		Literal literal = new Literal(string, datatype, language, createID());
		nodes.put(literal.id(), literal);
		return literal;
	}
}
