/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf;

import java.io.Serializable;

/**
 * A special (fast) index for accessing statements by their subject and predicate.
 * 
 * @author Michael Hedenus
 */
public class SPIndex implements Serializable
{
	private static final long serialVersionUID = 8626031299752232937L;

	/**
	 * The hash table.
	 */
	private static class Table implements Serializable
	{
		private static final long serialVersionUID = 4117623981725638763L;

		/**
		 * An entry of the hash table. Each entry may have more than one value
		 * In most cases is there is only one value, but consider e.g. two statements with different languages.
		 * If there are hash collisions the multiple entries stored as linked list.
		 */
		private static class Entry implements Serializable
		{
			private static final long serialVersionUID = -8662731581118894305L;

			private final SPKey key;
			private short count;
			private Statement[] statements;
			private Entry _next;

			public Entry(SPKey key, Statement statement)
			{
				this.key = key;
				this.count = 1;
				this.statements = new Statement[] { statement };
			}

			public void add(Statement statement)
			{
				final Statement[] oldValues = this.statements;
				final int oldCount = this.count;

				Statement[] newValues = new Statement[count + 1];
				for (int i = 0; i < oldCount; i++)
				{
					newValues[i] = oldValues[i];
				}

				newValues[count++] = statement;
				this.statements = newValues;
			}
		}

		/**
		 * The table's length (must be power of two).
		 */
		private final int length;

		/**
		 * Table's length -1;
		 */
		private final int mod;

		/**
		 * The table's load threshold.
		 */
		private final int threshold;

		/**
		 * The entries.
		 */
		private final Entry[] entries;

		public Table(int length, float loadFactor)
		{
			this.length = length;
			this.mod = length - 1;
			threshold = (int)(length * loadFactor);

			entries = new Entry[length];
		}

		public int indexFor(SPKey key)
		{
			return key.hashCode() & mod;
		}

		public int indexFor(String subject, String predicate)
		{
			return SPKey.hashCode(subject, predicate) & mod;
		}

		/**
		 * Adds a new key - value mapping to the table.
		 */
		public void add(SPKey key, Statement statement)
		{
			// find slot in table
			int index = indexFor(key);

			Entry entry = entries[index];
			if (entry == null)
			{
				entries[index] = new Entry(key, statement);
			}
			else
			{
				while (true)
				{
					// add vew to entry for same key --> done
					if (entry.key.equals(key))
					{
						entry.add(statement);
						break;
					}

					if (entry._next == null)
					{
						// no next entry --> append new entry and done
						entry._next = new Entry(key, statement);
						break;
					}
					else
					{
						// go next entry in chain
						entry = entry._next;
					}
				}
			}
		}

		/**
		 * Returns all values for the key (subject, predicate) or null.
		 */
		public Statement[] get(String subject, String predicate)
		{
			int index = indexFor(subject, predicate);
			for (Entry entry = entries[index]; entry != null; entry = entry._next)
			{
				if (entry.key.equals(subject, predicate))
				{
					return entry.statements;
				}
			}
			return null;
		}
	}

	private Table table;
	private int size;
	private float loadFactor = 0.9f;

	/**
	 * Creates a table with inital capacity of 128.
	 */
	public SPIndex()
	{
		this(128);
	}

	/**
	 * Creates a table with an inital capacity which must be a power of two.
	 */
	public SPIndex(int initialLength)
	{
		table = new Table(initialLength, loadFactor);
	}

	/**
	 * Returns the number of indexed statements.
	 */
	public int getSize()
	{
		return size;
	}

	public boolean isEmpty()
	{
		return size == 0;
	}

	public float getLoadFactor()
	{
		return loadFactor;
	}

	public void setLoadFactor(float loadFactor)
	{
		this.loadFactor = loadFactor;
	}

	/**
	 * Returns an array of Statements for the given subject and predicate or <code>null</code>
	 * if there is no statement.
	 * <br><b>Note</b>: The returned array is <i>live</i>, do not change it!
	 */
	public Statement[] get(String subject, String predicate)
	{
		return table.get(subject, predicate);
	}

	/**
	 * Adds a statement to the index. 
	 */
	public void add(Statement statement)
	{
		if (size >= table.threshold)
		{
			resize();
		}

		table.add(new SPKey(statement.getSubject().about(), statement.getPredicate()), statement);
		size++;
	}

	/**
	 * Creates a new table and copies all old values.
	 */
	private void resize()
	{
		System.out.println("SPIndex: before resize: " + getStatistics());

		Table oldTable = this.table;
		Table newTable = new Table(oldTable.length * 2, loadFactor);

		// for whole table
		for (int index = 0; index < oldTable.length; index++)
		{
			// for all entries
			for (Table.Entry e = oldTable.entries[index]; e != null; e = e._next)
			{
				// for each statement
				for (int j = 0; j < e.count; j++)
				{
					newTable.add(e.key, e.statements[j]);
				}
			}
		}

		this.table = newTable;

		System.out.println("SPIndex: after resize: " + getStatistics());
	}

	String getStatistics()
	{
		int count = 0;
		int empty = 0;
		int totalCollisions = 0;
		int maxCollisions = 0;

		for (int index = 0; index < table.length; index++)
		{
			Table.Entry e = table.entries[index];

			if (e != null)
			{
				int col = 0;
				count++;
				for (e = e._next; e != null; e = e._next)
				{
					count++;
					totalCollisions++;
					col++;
				}
				if (col > maxCollisions)
				{
					maxCollisions = col;
				}
			}
			else
			{
				empty++;
			}
		}

		return "size=" + table.length + //
				", count=" + count + " (" + (100 * count / table.length) + "%)" + //
				", empty=" + empty + " (" + (100 * empty / table.length) + "%)" + //
				", collisions=" + totalCollisions + " (" + (100 * totalCollisions / count) + "%, max=" + maxCollisions + ")" + //
				", scans=" + (1.0f * count / (table.length - empty));
	}
}
