/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf;

import java.io.Serializable;

/**
 * Subject-predicate key to quickly answer query <code>(subject, predicate, *)</code>.
 * 
 * @author Michael Hedenus
 */
public class SPKey implements Serializable
{
	private static final long serialVersionUID = -8926121222874730930L;

	private final String subject;
	private final String predicate;
	private final int hashCode;

	public SPKey(String subject, String predicate)
	{
		this.subject = subject;
		this.predicate = predicate;
		this.hashCode = hashCode(subject, predicate);
	}

	/**
	 * Computes a hash code for the <code>(subject, predicate)</code> pair.
	 */
	public static int hashCode(String subject, String predicate)
	{
		/*
		int hash = 5381;
		final int len1 = subject.length();
		for (int i = 0; i < len1; i++)
		{
			hash = (hash << 5) + hash + subject.charAt(i);
		}
		final int len2 = predicate.length();
		for (int i = 0; i < len2; i++)
		{
			hash = (hash << 5) + hash + predicate.charAt(i);
		}
		*/
		
		// String caches the hashCode, so the code must not be recomputed each time
		int hash = (65521 * subject.hashCode()) + predicate.hashCode();
		
		return secureHash(hash);
	}

	private static int secureHash(int hash)
	{
		// This function ensures that hashCodes that differ only by
		// constant multiples at each bit position have a bounded
		// number of collisions (approximately 8 at default load factor).
		hash ^= (hash >>> 20) ^ (hash >>> 12);
		return hash ^ (hash >>> 7) ^ (hash >>> 4);
	}

	public String getPredicate()
	{
		return predicate;
	}

	public String getSubject()
	{
		return subject;
	}

	public int hashCode()
	{
		return hashCode;
	}

	public boolean equals(Object obj)
	{
		SPKey that = (SPKey)obj;
		return this.hashCode == that.hashCode && //
				this.subject.equals(that.subject) && this.predicate.equals(that.predicate);
	}

	public boolean equals(String subject, String predicate)
	{
		return this.subject.equals(subject) && this.predicate.equals(predicate);
	}

	public String toString()
	{
		return "SPKey(subject=" + subject + ", predicate=" + predicate + ", hash=" + hashCode + ")";
	}
}
