/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf.serialize;

import java.io.PrintStream;

import de.hedenus.rdf.RDFException;

/**
 * A debug listener which print the statements to a print stream before it dispatches the
 * RDF events to the decorated listener.
 * 
 * @author Michael Hedenus
 */
public class DebugRDFEventListener implements RDFEventListener
{
	private final PrintStream ps;
	private final RDFEventListener nextListener;
	private int counter = 0;

	public DebugRDFEventListener()
	{
		this(System.out, null);
	}

	public DebugRDFEventListener(RDFEventListener nextListener)
	{
		this(System.out, nextListener);
	}

	public DebugRDFEventListener(PrintStream ps, RDFEventListener nextListener)
	{
		this.ps = ps;
		this.nextListener = nextListener;
	}

	public void newStatement(String subject, int subjectType, String predicate, String object, int objectType, String datatype, String lang)
	{
		ps.print(++counter);
		ps.print('\t');
		ps.print(toString(subject, subjectType, null, null));
		ps.print(' ');
		ps.print(toString(predicate, RDFEventListener.NODETYPE_URIREF, null, null));
		ps.print(' ');
		ps.print(toString(object, objectType, datatype, lang));
		ps.println(" .");
		ps.flush();

		if (nextListener != null)
		{
			nextListener.newStatement(subject, subjectType, predicate, object, objectType, datatype, lang);
		}
	}

	private String toString(String value, int type, String datatype, String lang)
	{
		switch (type)
		{
			case RDFEventListener.NODETYPE_BLANK_NODE:
				return "_:" + value;
			case RDFEventListener.NODETYPE_URIREF:
				return "<" + value + ">";
			case RDFEventListener.NODETYPE_LITERAL:
				return "\"" + value + "\"" + (lang != null ? "@" + lang : "") + (datatype != null ? "^^<" + datatype + ">" : "");
			default:
				throw new RDFException("Unknown node type: " + type);
		}
	}
}
