/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf.serialize;

import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsEnvironment;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Date;
import java.util.Iterator;

import de.hedenus.graph.Node;
import de.hedenus.rdf.BlankNode;
import de.hedenus.rdf.Literal;
import de.hedenus.rdf.RDFGraph;
import de.hedenus.rdf.Statement;
import de.hedenus.rdf.StringUtil;
import de.hedenus.rdf.URIRefNode;
import de.hedenus.www.NamespaceContext;

/**
 * A <a href="http://www.infosun.fim.uni-passau.de/Graphlet/GML/gml-tr.html"><b>G</b>raph <b>M</b>odelling <b>L</b>anguage</a> serializer.
 * 
 * @author Michael Hedenus
 */
public class GMLSerializer extends RDFSerializer
{
	private boolean showBlankNodeIds = true;

	public boolean isShowBlankNodeIds()
	{
		return showBlankNodeIds;
	}

	public void setShowBlankNodeIds(boolean showBlankNodeIds)
	{
		this.showBlankNodeIds = showBlankNodeIds;
	}

	public void serialize(RDFGraph graph, OutputStream outputStream) throws IOException
	{
		// create dummy image to construct graphics
		GraphicsEnvironment graphicsEnvironment = GraphicsEnvironment.getLocalGraphicsEnvironment();
		GraphicsConfiguration graphicsConfiguration = graphicsEnvironment.getDefaultScreenDevice().getDefaultConfiguration();
		BufferedImage bufferedImage = graphicsConfiguration.createCompatibleImage(1, 1);
		Font font = new Font("Dialog", Font.PLAIN, 12);
		FontMetrics fontMetrics = bufferedImage.createGraphics().getFontMetrics(font);
		int fontHeight = fontMetrics.getHeight();

		final NamespaceContext namespaceContext = graph.getNamespaceContext();
	
		PrintWriter pw = new PrintWriter(outputStream);
		pw.println("graph [");
		pw.println("comment \"Exported by " + getClass().getName() + " on " + new Date() + "\"");
		pw.println("directed 1");

		for (Iterator i = graph.nodes().iterator(); i.hasNext();)
		{
			Node node = (Node)i.next();

			String label = "";
			if (node instanceof BlankNode)
			{
				label = !showBlankNodeIds ? "" : node.label();
			}
			if (node instanceof URIRefNode)
			{
				String uri = ((URIRefNode)node).getURI();
				label = namespaceContext.toShortForm(uri);
			}
			if (node instanceof Literal)
			{
				Literal literal = (Literal)node;
				String datatype = literal.getDatatype();
				String lang = literal.getLanguage();
				label = "'" + StringUtil.normalizeSpace(literal.getString()) + "'" + (lang != null ? ("@" + lang) : "") + (datatype != null ? ("^^" + namespaceContext.toShortForm(datatype)) : "");
			}

			pw.println("node [");
			pw.println("id " + getId(node));
			pw.println("label \"" + label + "\"");
			pw.println("graphics [");

			int stringWidth = fontMetrics.stringWidth(label);
			int rows = 1;
			if (node instanceof URIRefNode)
			{
				pw.println("w " + (stringWidth + 16));
				pw.println("h " + (fontHeight + 8));
				pw.println("type \"oval\"");
				pw.println("fill \"#CCCCFF\"");
			}
			if (node instanceof BlankNode)
			{
				pw.println("w " + (stringWidth + 16));
				pw.println("h " + (fontHeight + 8));
				pw.println("type \"oval\"");
				pw.println("fill \"#FFFFFF\"");
			}
			if (node instanceof Literal)
			{
				//String text = ((Literal)node).getString();
				pw.println("w " + (stringWidth + 8));
				pw.println("h " + ((fontHeight * rows) + 4));
				pw.println("type \"rectangle\"");
				pw.println("fill \"#FFFF88\"");
			}
			pw.println("outline \"#000000\"");
			pw.println("]");
			pw.println("]");
		}

		for (Iterator i = graph.arcs().iterator(); i.hasNext();)
		{
			Statement statement = (Statement)i.next();
			
			String color = "000000";
			int statementType = graph.getStatementType(statement);
			if((statementType & RDFGraph.STATEMENT_INFERED) == RDFGraph.STATEMENT_INFERED)
			{
				color = "5555FF";
			}
			if((statementType & RDFGraph.STATEMENT_AXIOM) == RDFGraph.STATEMENT_AXIOM)
			{
				color = "FFAAAA";
			}
			
			pw.println("edge [");
			pw.println("label \"" + namespaceContext.toShortForm(statement.getPredicate()) + "\"");
			pw.println("source " + getId(statement.getSubject()));
			pw.println("target " + getId(statement.getObject()));
			pw.println("graphics [");
			pw.println("arrow \"last\"");
			pw.print("fill \"#");
			pw.print(color);
			pw.println("\"");
			pw.println("]");
			pw.println("]");
		}

		pw.println("]");
		pw.flush();
	}

	private String getId(Node node)
	{
		return node.id().substring(2);
	}
}
