/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf.serialize;

import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import de.hedenus.mobilexml.Attribute;
import de.hedenus.mobilexml.CharBuffer;
import de.hedenus.mobilexml.Element;
import de.hedenus.mobilexml.XMLException;
import de.hedenus.mobilexml.XMLReader;
import de.hedenus.mobilexml.XMLReaderEventListener;
import de.hedenus.mobilexml.XMLReader.Scope;
import de.hedenus.rdf.RDFConstants;

/**
 * A RDF-XML parser.
 * 
 * @author Michael Hedenus
 */
public class RDFXMLParser extends RDFParser implements XMLReaderEventListener
{
	private static final int POS_DOCUMENT = -1;
	private static final int POS_RDF_ROOT_ELEMENT = 0;
	private static final int POS_NODE_ELEMENT = 1;
	private static final int POS_PROPERTY_ELEMENT = 2;

	private final HashMap parseTypes = new HashMap();
	private final ParseType defaultParseType = new DefaultParseType();
	private final ResourceParseType resourceParseType = new ResourceParseType();
	private final CollectionParseType collectionParseType = new CollectionParseType();
	private final LiteralParseType literalParseType = new LiteralParseType();

	private RDFEventListener rdfEventListener;

	private static final String blankNodeURI = "http://www.hedenus.de/java/RDFXMLParser#Blank";
	private int blankNodeCounter = 0;

	/**
	 * The XMLReader instance.
	 */
	private final XMLReader xmlReader = new XMLReader(this)
	{
		protected Scope createScope(Scope parent)
		{
			return new RDFScope(parent);
		}
	};

	public RDFXMLParser()
	{
		this.xmlReader.setIgnoreEmptyText(true);
	}

	private int getLineNumber()
	{
		return xmlReader.getLineNumber();
	}

	public void parse(RDFEventListener rdfEventListener, InputStream in, String encoding, String documentBase) throws IOException, XMLException
	{
		this.rdfEventListener = rdfEventListener;
		this.xmlReader.read(in, encoding, documentBase);

		// clean up
		this.rdfEventListener = null;
	}

	//
	// the events are delegated to the various parseTypes
	//

	public void elementBegin(Element element, Scope s) throws XMLException
	{
		RDFScope scope = (RDFScope)s;
		RDFScope parentScope = (RDFScope)scope.getParent();
		scope.parseType = parentScope.parseType.getDescendantParseType();

		scope.parseType.elementBegin(element, scope);
	}

	public void elementEnd(Element element, Scope s) throws XMLException
	{
		RDFScope scope = (RDFScope)s;
		RDFScope parentScope = (RDFScope)scope.getParent();
		scope.parseType = parentScope.parseType.getDescendantParseType();
		scope.parseType.elementEnd(element, scope);
	}

	public void text(String text, Scope s) throws XMLException
	{
		RDFScope scope = (RDFScope)s;
		scope.parseType.text(text, scope);
	}

	private class RDFScope extends XMLReader.Scope
	{
		int pos = POS_DOCUMENT;
		ParseType parseType = defaultParseType;
		int nodeType = RDFEventListener.NODETYPE_BLANK_NODE;
		int liCounter = 0;
		String about = null;
		String elementURI = null;
		String text = null;
		String datatype = null;
		String list = null;
		boolean xmlLiteral = false;

		//list of property attributes
		ArrayList propertyAttributes = new ArrayList();

		protected RDFScope(Scope parent)
		{
			super(parent);
		}

		protected void reset()
		{
			super.reset();

			pos = POS_DOCUMENT;
			parseType = null;
			nodeType = RDFEventListener.NODETYPE_BLANK_NODE;
			liCounter = 0;
			about = null;
			elementURI = null;
			text = null;
			datatype = null;
			list = null;
			xmlLiteral = false;
			propertyAttributes.clear();
		}
	}

	/**
	 * Handler base for the various rdf:parseTypes.
	 */
	private abstract class ParseType implements XMLReaderEventListener
	{
		private final String parseType;

		protected ParseType(String parseType)
		{
			this.parseType = parseType;
			parseTypes.put(parseType, this);
		}

		public String toString()
		{
			return parseType;
		}

		/**
		 * Returns the parse type that follows this parse type.
		 */
		public abstract ParseType getDescendantParseType();
	}

	/**
	 * Default parseType.
	 */
	private class DefaultParseType extends ParseType
	{
		protected DefaultParseType()
		{
			this("Default");
		}

		public ParseType getDescendantParseType()
		{
			return defaultParseType;
		}

		/**
		 * Used by subclass.
		 */
		protected DefaultParseType(String parseType)
		{
			super(parseType);
		}

		public void elementBegin(Element element, Scope s) throws XMLException
		{
			RDFScope scope = (RDFScope)s;
			RDFScope parentScope = (RDFScope)scope.getParent();

			switch (parentScope.pos)
			{
				case POS_DOCUMENT:
					scope.pos = POS_RDF_ROOT_ELEMENT;
					// check RDF root element
					if (element.equals(RDFConstants.RDF, RDFConstants.NS))
					{
						//System.out.println("---[RDF]---");
					}
					break;

				case POS_RDF_ROOT_ELEMENT:
					scope.pos = POS_NODE_ELEMENT;
					break;

				case POS_NODE_ELEMENT:
					scope.pos = POS_PROPERTY_ELEMENT;
					break;

				case POS_PROPERTY_ELEMENT:
					scope.pos = POS_NODE_ELEMENT;
					break;

				default:
					break;
			}

			switch (scope.pos)
			{
				case POS_NODE_ELEMENT:
					parseNodeElement(element, scope);
					break;
				case POS_PROPERTY_ELEMENT:
					parsePropertyElement(element, scope);
					break;
				default:
					break;
			}
		}

		public void text(String text, Scope s) throws XMLException
		{
			RDFScope scope = (RDFScope)s;
			switch (scope.pos)
			{
				case POS_PROPERTY_ELEMENT:
					scope.text = text;
					break;

				default:
					break;
			}
		}

		public void elementEnd(Element element, Scope s) throws XMLException
		{
			RDFScope scope = (RDFScope)s;
			RDFScope parentScope = (RDFScope)scope.getParent();

			switch (scope.pos)
			{
				case POS_PROPERTY_ELEMENT:

					if (scope.xmlLiteral)
					{
						if (scope.about == null)
						{
							rdfEventListener.newStatement( //
									parentScope.about, parentScope.nodeType, //
									scope.elementURI, //
									literalParseType.buffer.toString(), //
									RDFEventListener.NODETYPE_LITERAL, RDFConstants.XMLLITERAL_URI, null);
						}
					}
					//
					// if the content of a property element is only text then this is a Literal.
					// 
					else if (scope.getContent() == Scope.CONTENT_TEXT || scope.getContent() == Scope.CONTENT_EMPTY)
					{
						if (scope.about == null)
						{
							rdfEventListener.newStatement( //
									parentScope.about, parentScope.nodeType, //
									scope.elementURI, //
									scope.text == null ? "" : scope.text, //
									RDFEventListener.NODETYPE_LITERAL, scope.datatype, scope.datatype == null ? scope.getLang() : null);
						}
					}

					//
					// close rdf:List
					//
					if (scope.list != null)
					{
						rdfEventListener.newStatement( //
								scope.list, RDFEventListener.NODETYPE_BLANK_NODE, //
								RDFConstants.REST_URI, //
								RDFConstants.NIL_URI, RDFEventListener.NODETYPE_URIREF, null, null);
					}

					break;

				default:
					break;
			}
		}

		protected void parseAttributes(Element element, RDFScope scope) throws XMLException
		{
			//
			// if not rdf:Description: element URI = rdf:type
			//
			if (!element.equals(RDFConstants.DESCRIPTION, RDFConstants.NS) && !element.equals(RDFConstants.RDF, RDFConstants.NS))
			{
				scope.elementURI = element.getItemURI();
			}

			final Attribute[] attrs = scope.getAttributes();
			for (int i = 0; i < attrs.length; i++)
			{
				Attribute attr = attrs[i];

				if ((scope.pos == POS_NODE_ELEMENT && attr.equals("about", RDFConstants.NS)) || (scope.pos == POS_PROPERTY_ELEMENT && attr.equals("resource", RDFConstants.NS)))
				{
					//
					// either rdf:about (subject) or rdf:resource (ojbect)
					//
					scope.nodeType = RDFEventListener.NODETYPE_URIREF;

					String uri = attr.getValue();
					if (startsWithHash(uri))
					{
						// uri is a fragement ref
						scope.about = resolveFragmentIdentifier(scope, uri);
					}
					else
					{
						scope.about = scope.resolveURI(uri);
					}
				}
				else if (scope.pos == POS_NODE_ELEMENT && attr.equals("ID", RDFConstants.NS))
				{
					//
					// rdf:ID - definition of a URI reference = xml:base + "#" + ID
					//
					scope.nodeType = RDFEventListener.NODETYPE_URIREF;
					scope.about = resolveFragmentIdentifier(scope, attr.getValue());
				}
				else if (attr.equals(RDFConstants.NODE_ID, RDFConstants.NS))
				{
					//
					// rdf:nodeID - blank node identifier
					//
					scope.nodeType = RDFEventListener.NODETYPE_BLANK_NODE;
					scope.about = attr.getValue();
				}
				else if (attr.equals(RDFConstants.DATATYPE, RDFConstants.NS))
				{
					//
					// rdf:datatype - literal's datatype
					//
					scope.datatype = scope.resolveURI(attr.getValue());
				}
				else if (attr.equals(RDFConstants.PARSETYPE, RDFConstants.NS))
				{
					if (scope.pos != POS_PROPERTY_ELEMENT)
					{
						throw new RDFXMLException("Attribute 'parseType' is only allowed for property elements.", getLineNumber());
					}

					//
					// rdf:parseType
					//
					scope.parseType = (ParseType)parseTypes.get(attr.getValue());
					if (scope.parseType == null)
					{
						throw new RDFXMLException("Unsupported parse type '" + attr.getValue() + "'.", getLineNumber());
					}

					if (scope.parseType == resourceParseType)
					{
						// the element does behave as if it were a rdf:Description node-element
						scope.about = generateBlankNodeId();
						scope.nodeType = RDFEventListener.NODETYPE_BLANK_NODE;
						scope.pos = POS_NODE_ELEMENT;
					}

					if (scope.parseType == literalParseType)
					{
						scope.datatype = RDFConstants.XMLLITERAL_URI;
						scope.xmlLiteral = true;
						literalParseType.reset();
					}
				}
				else
				{
					//
					// not a syntax name: property attribute
					//
					scope.propertyAttributes.add(attr);
				}
			}
		}

		protected void parseNodeElement(Element element, RDFScope scope) throws XMLException
		{
			RDFScope parentScope = (RDFScope)scope.getParent();
			RDFScope greatParentScope = (RDFScope)parentScope.getParent();

			// attributes
			parseAttributes(element, scope);

			//create temporary identifier for blanknode
			if (scope.nodeType == RDFEventListener.NODETYPE_BLANK_NODE && scope.about == null)
			{
				scope.about = generateBlankNodeId();
			}

			if (scope.elementURI != null)
			{
				//
				// an element type is interpreted as rdf:type statement
				// <mh:A about="x"/> means: x rdf:type mh:A
				// same declaration in programming languages, e.g. Java: mh.A x = new mh.A() 
				//
				rdfEventListener.newStatement( //
						scope.about, scope.nodeType, //
						RDFConstants.TYPE_URI, //
						scope.elementURI, RDFEventListener.NODETYPE_URIREF, null, null);
			}

			//
			// other properties
			//
			processPropertyAttributes(element, scope);

			//
			// attach to ancestors
			//
			if (greatParentScope.about != null)
			{
				if (parentScope.parseType == collectionParseType)
				{
					//
					// generate rdf:List
					//
					String list = generateBlankNodeId();

					// list rdf:type rdf:List
					rdfEventListener.newStatement( //
							list, RDFEventListener.NODETYPE_BLANK_NODE, //
							RDFConstants.TYPE_URI, //
							RDFConstants.LIST_URI, RDFEventListener.NODETYPE_URIREF, null, null);

					// list rdf:first current
					rdfEventListener.newStatement( //
							list, RDFEventListener.NODETYPE_BLANK_NODE, //
							RDFConstants.FIRST_URI, //
							scope.about, scope.nodeType, null, null);

					if (parentScope.list == null)
					{
						rdfEventListener.newStatement( //
								greatParentScope.about, greatParentScope.nodeType, //
								parentScope.elementURI, //
								list, RDFEventListener.NODETYPE_BLANK_NODE, null, null);
					}
					else
					{
						rdfEventListener.newStatement( //
								parentScope.list, RDFEventListener.NODETYPE_BLANK_NODE, //
								RDFConstants.REST_URI, //
								list, RDFEventListener.NODETYPE_BLANK_NODE, null, null);
					}

					parentScope.list = list;
				}
				else
				{
					// <subject>
					//   <predicate>
					//      <object>
					rdfEventListener.newStatement( //
							greatParentScope.about, greatParentScope.nodeType, //
							parentScope.elementURI, //
							scope.about, scope.nodeType, null, null);
				}
			}
		}

		protected void processPropertyAttributes(Element element, RDFScope scope) throws XMLException
		{
			for (int i = 0; i < scope.propertyAttributes.size(); i++)
			{
				Attribute propertyAttr = (Attribute)scope.propertyAttributes.get(i);

				rdfEventListener.newStatement( //
						scope.about, scope.nodeType, //
						propertyAttr.getItemURI(), //
						propertyAttr.getValue(), RDFEventListener.NODETYPE_LITERAL, null, scope.getLang());
			}
		}

		protected void parsePropertyElement(Element element, RDFScope scope) throws XMLException
		{
			parseAttributes(element, scope);

			if (scope.elementURI != null)
			{
				RDFScope parentScope = (RDFScope)scope.getParent();

				//
				// replace rdf:li by rdf:_<n>
				//
				if (scope.elementURI.equals(RDFConstants.LI_URI))
				{
					parentScope.liCounter++;
					scope.elementURI = RDFConstants.NS + "_" + parentScope.liCounter;
				}

				if (parentScope.about != null && scope.about != null)
				{
					rdfEventListener.newStatement(//
							parentScope.about, parentScope.nodeType, //
							scope.elementURI, //
							scope.about, scope.nodeType, null, null);
				}
			}
			else
			{
				throw new RDFXMLException("Element type missing.", getLineNumber());
			}
		}
	}

	/**
	 * The "Resource" parseType my be applied to property nodes and simulates a blank node.
	 */
	private class ResourceParseType extends DefaultParseType
	{
		protected ResourceParseType()
		{
			super("Resource");
		}
	}

	/**
	 * The "Collection" parseType adds list structure.
	 */
	private class CollectionParseType extends DefaultParseType
	{
		protected CollectionParseType()
		{
			super("Collection");
		}
	}

	private class LiteralParseType extends ParseType
	{
		private CharBuffer buffer = new CharBuffer(128);
		private boolean root = true;

		protected LiteralParseType()
		{
			super("Literal");
		}
		
		public void reset()
		{
			buffer.reset();
			root = true;
		}

		public ParseType getDescendantParseType()
		{
			return this;
		}

		public void elementBegin(Element element, Scope scope) throws XMLException
		{
			buffer.append('<').append(element.getName());

			if (root)
			{
				root = false;

				String defaultns = scope.getDefaultNamespace();
				if (defaultns != null)
				{
					buffer.append(" xmlns=\"").append(defaultns).append('\"');
				}
			}

			HashMap namespaceMap = scope.getNamespaceMap();
			if (!namespaceMap.isEmpty())
			{
				for (Iterator i = namespaceMap.keySet().iterator(); i.hasNext();)
				{
					String prefix = (String)i.next();
					String nsuri = (String)namespaceMap.get(prefix);
					buffer.append(" xmlns:").append(prefix).append('=').append('\"').append(nsuri).append('\"');
				}
			}

			buffer.append('>');
		}

		public void elementEnd(Element element, Scope scope) throws XMLException
		{
			buffer.append('<').append('/').append(element.getName()).append('>');
		}

		public void text(String text, Scope scope) throws XMLException
		{
			buffer.append(text);
		}
	}

	/**
	 * Generates a new blank node id.
	 */
	private String generateBlankNodeId()
	{
		return blankNodeURI + (++blankNodeCounter);
	}

	/**
	 * Resolves a fragment identifier againt xml:base in scope.
	 */
	private String resolveFragmentIdentifier(Scope scope, String ref)
	{
		String base = scope.getBase();
		if (base == null)
		{
			base = "";
		}

		// try to avoid double '#'
		final boolean baseEndsWithHash = base.length() > 0 && base.charAt(base.length() - 1) == '#';
		final boolean refStartsWithHash = startsWithHash(ref);

		if (baseEndsWithHash && refStartsWithHash)
		{
			return base + ref.substring(1);
		}
		else if (baseEndsWithHash ^ refStartsWithHash)
		{
			return base + ref;
		}
		else
		{
			return base + "#" + ref;
		}
	}

	private boolean startsWithHash(String uri)
	{
		return uri.length() > 0 && uri.charAt(0) == '#';
	}
}