/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdf.serialize;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.util.Date;
import java.util.Iterator;
import java.util.TreeSet;

import de.hedenus.rdf.BlankNode;
import de.hedenus.rdf.Literal;
import de.hedenus.rdf.RDFConstants;
import de.hedenus.rdf.RDFGraph;
import de.hedenus.rdf.RDFNode;
import de.hedenus.rdf.URIRefNode;
import de.hedenus.rdf.Statement;
import de.hedenus.www.NamespaceContext;

/**
 * RDF-XML serializer.
 * 
 * @author Michael Hedenus
 */
public class RDFXMLSerializer extends RDFSerializer
{
	private NamespaceContext namespaceContext;
	private TreeSet tempPrefixes;
	private PrintWriter pw;

	public void serialize(RDFGraph graph, OutputStream outputStream) throws IOException
	{
		this.namespaceContext = graph.getNamespaceContext().copy();
		this.tempPrefixes = new TreeSet();

		this.pw = new PrintWriter(outputStream);

		//
		// prolog
		//
		pw.println("<?xml version=\"1.0\" encoding=\"US-ASCII\"?>");
		pw.print("<!-- created ");
		pw.print(new Date());
		pw.println("-->");

		//
		// DTD
		//
		pw.print("<!DOCTYPE ");
		pw.print(qname(RDFConstants.RDF, RDFConstants.NS));
		pw.println(" [");
		for (Iterator i = namespaceContext.getPrefixes(); i.hasNext();)
		{
			String prefix = (String)i.next();
			if (!"".equals(prefix))
			{
				String uri = namespaceContext.getNamespaceURI(prefix);
				pw.print("\t<!ENTITY ");
				pw.print(prefix);
				pw.print(" \"");
				pw.print(uri);
				pw.println("\">");
			}
		}
		pw.println("]>");

		//
		// root element
		//
		pw.print("<");
		pw.println(qname(RDFConstants.RDF, RDFConstants.NS));
		for (Iterator i = namespaceContext.getPrefixes(); i.hasNext();)
		{
			String prefix = (String)i.next();
			if (!"".equals(prefix))
			{
				pw.print("\txmlns:");
				pw.print(prefix);
				pw.print("=\"&");
				pw.print(prefix);
				pw.println(";\"");
			}
			else
			{
				pw.print("\txmlns=\"");
				pw.print(namespaceContext.getNamespaceURI(prefix));
				pw.println("\"");
			}
		}
		pw.println("\t>");

		//
		// statements
		//
		RDFNode lastSubject = null;

		TreeSet statements = new TreeSet(graph.arcs());
		for (Iterator i = statements.iterator(); i.hasNext();)
		{
			Statement statement = (Statement)i.next();
			RDFNode subject = statement.getSubject();
			String predicate = statement.getPredicate();
			RDFNode object = statement.getObject();

			// group descriptions
			if (lastSubject == null || !subject.equals(lastSubject))
			{
				if (lastSubject != null)
				{
					pw.print("\t</");
					pw.print(qname(RDFConstants.DESCRIPTION, RDFConstants.NS));
					pw.println(">");
				}

				pw.println();
				pw.print("\t<");
				pw.print(qname(RDFConstants.DESCRIPTION, RDFConstants.NS));
				if (subject instanceof URIRefNode)
				{
					printAttribute(RDFConstants.ABOUT, RDFConstants.NS, ((URIRefNode)subject).getURI());
				}
				if (subject instanceof BlankNode)
				{
					printAttribute(RDFConstants.NODE_ID, RDFConstants.NS, subject.id());
				}
				pw.println(">");

				lastSubject = subject;
			}

			pw.print("\t\t<");
			String label = toShortForm(predicate);
			pw.print(label);
			
			// if temp prefix: write temp namespace
			int ci = label.indexOf(':');
			String prefix = label.substring(0, ci);
			if(tempPrefixes.contains(prefix))
			{
				pw.print(" xmlns:");
				pw.print(prefix);
				pw.print("=\"");
				pw.print(namespaceContext.getNamespaceURI(prefix));
				pw.print("\"");
			}
			
			if (object instanceof Literal)
			{
				Literal literal = (Literal)object;
				if (literal.getDatatype() != null)
				{
					printAttribute(RDFConstants.DATATYPE, RDFConstants.NS, literal.getDatatype());
				}
				if (literal.getLanguage() != null)
				{
					pw.print(" xml:lang=\"");
					pw.print(literal.getLanguage());
					pw.print("\"");
				}
				pw.print(">");
				pw.print(xmlString(literal.getString()));
				pw.print("</");
				pw.print(toShortForm(predicate));
				pw.println(">");
			}
			else if (object instanceof URIRefNode)
			{
				printAttribute(RDFConstants.RESOURCE, RDFConstants.NS, ((URIRefNode)object).getURI());
				pw.println("/>");
			}
			else if (object instanceof BlankNode)
			{
				printAttribute(RDFConstants.NODE_ID, RDFConstants.NS, object.id());
				pw.println("/>");
			}
		}

		// end last subject
		if (lastSubject != null)
		{
			pw.print("\t</");
			pw.print(qname(RDFConstants.DESCRIPTION, RDFConstants.NS));
			pw.println(">");
		}

		// end root
		pw.print("</");
		pw.print(qname(RDFConstants.RDF, RDFConstants.NS));
		pw.println(">");

		// cleanup
		pw.flush();
		pw = null;
		namespaceContext = null;
		tempPrefixes = null;
	}

	private void printAttribute(String localName, String namespaceURI, String value)
	{
		pw.print(" ");
		pw.print(qname(localName, namespaceURI));
		pw.print("=\"");
		pw.print(xmlString(value));
		pw.print("\"");
	}

	private final StringBuffer buffer = new StringBuffer();

	private String xmlString(String value)
	{
		buffer.setLength(0);
		for (int i = 0; i < value.length(); i++)
		{
			char c = value.charAt(i);
			if (c < 127)
			{
				switch (c)
				{
					case '\'':
						buffer.append("&apos;");
						break;

					case '\"':
						buffer.append("&quot;");
						break;

					case '<':
						buffer.append("&lt;");
						break;

					case '>':
						buffer.append("&gt;");
						break;

					case '&':
						buffer.append("&amp;");
						break;

					default:
						buffer.append(c);
						break;
				}
			}
			else
			{
				buffer.append("&#").append((int)c).append(';');
			}
		}

		return buffer.toString();
	}

	private String qname(String localName, String namespaceURI)
	{
		return namespaceContext.createQName(localName, namespaceURI);
	}

	private String toShortForm(String uri)
	{
		String label = namespaceContext.toShortForm(uri);
		if (uri == label)
		{
			// no prefix: auto create one
			String autoPrefix = namespaceContext.createAndDefineNamespacePrefix(uri);
			tempPrefixes.add(autoPrefix);
			label = namespaceContext.toShortForm(uri);
		}

		return label;
	}
}
