/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdfschema;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.TreeSet;

/**
 * Key - value mapping.
 * 
 * @author Michael Hedenus
 */
public class Mapping
{
	private class Entry implements Comparable
	{
		private final String key;
		private final String value;

		private Entry(String key, String value)
		{
			this.key = key;
			this.value = value;
		}

		public String toString()
		{
			return key + " " + value;
		}

		public int compareTo(Object that)
		{
			return this.key.compareTo(((Entry)that).key);
		}
	}

	private final HashMap mappings = new HashMap();

	/**
	 * Get the value of of the mapping for the specified key.
	 * @param key
	 * @return the value of the mapping or null if no value is found
	 */
	public String get(String key)
	{
		Entry entry = (Entry)mappings.get(key);
		if (entry != null)
		{
			return entry.value;
		}
		else
		{
			return null;
		}
	}

	/**
	 * Adds a new key-value mapping. If there is already a mapping with the same key, the first mapping takes precedence.
	 * @param key
	 * @param value
	 */
	public void add(String key, String value)
	{
		if (!mappings.containsKey(key))
		{
			mappings.put(key, new Entry(key, value));
		}
	}

	public void write(PrintStream ps)
	{
		TreeSet sorted = new TreeSet(mappings.values());
		for (Iterator i = sorted.iterator(); i.hasNext();)
		{
			ps.println(i.next());
		}
	}

	public void dump()
	{
		write(System.out);
	}

	/**
	 * Adds a mapping from an URL.
	 * File format is a list key-value pairs separeted by whitespace.
	 * There is one mapping per line. Use ';' to mark a line as a comment line.  
	 * @param url
	 * @throws IOException
	 */
	public void add(URL url) throws IOException
	{
		BufferedReader br = null;
		try
		{
			br = new BufferedReader(new InputStreamReader(url.openStream()));
			for (String line = null; br.ready() && (line = br.readLine()) != null;)
			{
				if (line.startsWith(";"))
				{
					// ignore comment
				}
				else
				{
					StringTokenizer tokens = new StringTokenizer(line);
					if (tokens.countTokens() >= 2)
					{
						String key = tokens.nextToken();
						String value = tokens.nextToken();
						this.add(key, value);
					}
				}
			}
		}
		finally
		{
			if (br != null)
			{
				br.close();
			}
		}
	}
}
