/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdfschema;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Locale;

import de.hedenus.mobilexml.XMLException;
import de.hedenus.rdf.Literal;
import de.hedenus.rdf.RDFConstants;
import de.hedenus.rdf.RDFException;
import de.hedenus.rdf.RDFGraph;
import de.hedenus.rdf.RDFNode;
import de.hedenus.rdf.Statement;
import de.hedenus.rdf.URIRefNode;
import de.hedenus.rdf.serialize.GraphBuilder;
import de.hedenus.rdf.serialize.RDFXMLParser;
import de.hedenus.xmltypes.TypeException;
import de.hedenus.xmltypes.TypeFactory;
import de.hedenus.xmltypes.TypeFactoryConfigurationException;

/**
 * Main interface for RDF databinding.
 * 
 * @author Michael Hedenus
 */
public class RDFDatabinding
{
	private final RDFGraph graph = new RDFGraph();
	private final RDFXMLParser parser = new RDFXMLParser();
	private final TypeFactory typeFactory;
	private final Mapping typemapping = new Mapping();

	public RDFDatabinding()
	{
		try
		{
			this.typeFactory = TypeFactory.newInstance();
		}
		catch (TypeFactoryConfigurationException e)
		{
			throw new RDFException("Cannot create type factory: " + e.getMessage() + ".");
		}

		//
		// default RDF Schema mapping
		//
		typemapping.add(Resource.URI, ResourceImpl.class.getName());
		typemapping.add(Container.URI, ContainerImpl.class.getName());
		typemapping.add(Alt.URI, AltImpl.class.getName());
		typemapping.add(Bag.URI, BagImpl.class.getName());
		typemapping.add(Seq.URI, SeqImpl.class.getName());
		typemapping.add(List.URI, ListImpl.class.getName());

		loadMappings(Mapping.class.getClassLoader());

		//typemapping.dump();
	}

	protected Mapping getTypeMapping()
	{
		return typemapping;
	}

	/**
	 * Loads all typemapping found in the classpath. The typemappings are specified in
	 * a textfile with the resouce name <code>rdfschema/typemapping</code>.
	 * @see de.hedenus.rdfschema.Mapping
	 * @param classLoader
	 */
	private void loadMappings(ClassLoader classLoader)
	{
		Enumeration enumeration = null;
		try
		{
			enumeration = classLoader.getResources("rdfschema/typemapping");
		}
		catch (IOException e1)
		{
			e1.printStackTrace();
		}
		if (enumeration != null)
		{
			while (enumeration.hasMoreElements())
			{
				URL url = (URL)enumeration.nextElement();
				try
				{
					typemapping.add(url);
				}
				catch (IOException e2)
				{
					e2.printStackTrace();
				}
			}
		}
	}

	public RDFGraph getGraph()
	{
		return graph;
	}

	public TypeFactory getTypeFactory()
	{
		return typeFactory;
	}

	/**
	 * For importing basic schemas.
	 * @param url
	 * @param statementType
	 * @throws IOException
	 */
	protected void parse(URL url, int statementType) throws IOException
	{
		parser.parse(new GraphBuilder(graph, statementType), url);
	}

	public void parse(URL url) throws IOException, XMLException
	{
		parser.parse(new GraphBuilder(graph), url);
	}

	public void parse(File file, String encoding) throws IOException, XMLException
	{
		parser.parse(new GraphBuilder(graph), file, encoding);
	}

	public void parse(InputStream in, String encoding, String documentBase) throws IOException, XMLException
	{
		parser.parse(new GraphBuilder(graph), in, encoding, documentBase);
	}

	/**
	 * Returns the resource identified by <code>about</code> or null if
	 * thers no such resource.
	 */
	public Resource getResource(String about)
	{
		RDFNode node = getGraph().getNode(about);
		if (node == null)
		{
			// no node found: no can do
			return null;
		}
		else
		{
			return getResource(node);
		}
	}

	private final HashMap boundResources = new HashMap();

	/**
	 * Returns or lazily creates a resource object for the RDF node.
	 */
	private Resource getResource(RDFNode node)
	{
		Resource resource = (Resource)boundResources.get(node);
		if (resource == null)
		{
			resource = createResourceImpl(node);
			boundResources.put(node, resource);
		}
		return resource;
	}

	/**
	 * Creates an implementation instance for a RDF resource.
	 */
	private Resource createResourceImpl(RDFNode node)
	{
		ResourceImpl resourceImpl;
		if (node instanceof de.hedenus.rdf.Literal)
		{
			resourceImpl = new LiteralImpl();
		}
		else
		{
			//
			// try to find type
			//
			String className = null;

			int resultCount = getGraph().queryStatement(node.about(), RDFConstants.TYPE_URI, "*", tempQueryResult);
			for (int i = 0; i < resultCount && className == null; i++)
			{
				Statement statement = (Statement)tempQueryResult.get(i);

				// the type must be resource
				RDFNode obj = statement.getObject();
				if (obj instanceof URIRefNode)
				{
					String typeURI = ((URIRefNode)obj).getURI();
					className = typemapping.get(typeURI);
				}
			}
			tempQueryResult.clear();

			if (className == null)
			{
				// no type mapping? fall back to resource
				resourceImpl = new ResourceImpl();
			}
			else
			{
				try
				{
					resourceImpl = (ResourceImpl)java.lang.Class.forName(className).newInstance();
				}
				catch (Exception e)
				{
					throw new RDFException("Cannot instantiate implementation of Resource: " + className + ".");
				}
			}
		}

		//
		// init 
		//
		resourceImpl.init(this, node);

		// done
		return resourceImpl;
	}

	/**
	 * Returns the value of <code>resource</code>'s property <code>propertyURI</code> interpreted 
	 * as datatype specified by <code>datatypeURI</code>.
	 * If the value is not a RDF literal the result will be a <code>de.hedenus.rdfschema.Resource</code>.
	 * @param resource
	 * @param propertyURI
	 * @param datatypeURI
	 * @return the property value
	 */
	public Object getPropertyValue(String resource, String propertyURI, String datatypeURI)
	{
		Object value = null;

		Statement statement = querySingleStatement(resource, propertyURI, "*");
		if (statement != null)
		{
			RDFNode objectNode = statement.getObject();
			value = getValueOfNode(objectNode, datatypeURI);
		}

		return value;
	}

	private Object getValueOfNode(RDFNode objectNode, String datatypeURI)
	{
		Object value = null;

		if (objectNode instanceof de.hedenus.rdf.Literal)
		{
			de.hedenus.rdf.Literal literal = (de.hedenus.rdf.Literal)objectNode;

			// if the datatype of the literal instance is set
			// then it must fit the specified datatype
			//
			// this allows interpreting a literal with a datatype
			// even if the literal instance has no datatype
			// BUT the other way round is illegal
			//
			//TODO check datatype hierarchy

			if (literal.getDatatype() != null && !literal.getDatatype().equals(datatypeURI))
			{
				throw new RDFException("Literal's datatype " + literal.getDatatype() + " does not match " + datatypeURI + ".");
			}

			if (datatypeURI != null)
			{
				try
				{
					value = getTypeFactory().getType(datatypeURI).parse(literal.getString());
				}
				catch (TypeException e)
				{
					throw new RDFException("Problem with datatype " + datatypeURI + ": " + e.getMessage());
				}
			}
			else
			{
				value = literal.getString();
			}
		}
		else
		{
			//
			// not a Literal: return the bound object
			//
			value = getResource(objectNode);
		}

		return value;
	}

	private ArrayList tempQueryResult = new ArrayList();

	public Resource[] getPropertyValues(String resource, String propertyURI)
	{
		int resultCount = getGraph().queryStatement(resource, propertyURI, "*", tempQueryResult);

		Resource[] result = new Resource[resultCount];
		for (int i = 0; i < resultCount; i++)
		{
			Statement statement = (Statement)tempQueryResult.get(i);
			RDFNode objectNode = statement.getObject();
			result[i] = getResource(objectNode);
		}

		tempQueryResult.clear();
		return result;
	}

	public Resource getPropertyValue(String resource, String propertyURI)
	{
		Resource value = null;

		Statement statement = querySingleStatement(resource, propertyURI, "*");
		if (statement != null)
		{
			RDFNode objectNode = statement.getObject();
			value = getResource(objectNode);
		}

		return value;
	}

	private Statement querySingleStatement(String subject, String predicate, String object)
	{
		int resultCount = getGraph().queryStatement(subject, predicate, object, tempQueryResult);

		Statement result = null;
		if (resultCount > 0)
		{
			result = (Statement)tempQueryResult.get(0);

			if (resultCount > 1)
			{
				String query = subject + " " + predicate + " " + object;
				System.err.println("Warning: ambigous query result " + query);
			}
		}

		tempQueryResult.clear();
		return result;
	}

	public String getPropertyValue(String resource, String propertyURI, Locale locale)
	{
		int resultCount = getGraph().queryStatement(resource, propertyURI, "*", tempQueryResult);

		String result = null;
		for (int i = 0; i < resultCount; i++)
		{
			RDFNode objectNode = ((Statement)tempQueryResult.get(i)).getObject();
			if (objectNode instanceof Literal)
			{
				Literal literal = (Literal)objectNode;

				// no locale specified: take the first
				// locale: try to find match
				if (locale == null || (locale.getLanguage().equals(literal.getLanguage())))
				{
					result = literal.getString();
					break;
				}
			}
		}

		tempQueryResult.clear();
		return result;
	}
}
