
/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.rdfschema;

import java.util.Locale;

import de.hedenus.www.Namespaces;

/**
 * This interface represents <code>rdfs:Resource</code>.
 * <br><b>Note:</b> Do not confuse with <code>de.hedenus.rdf.Resource</code>.
 * 
 * @author Michael Hedenus
 */
public interface Resource
{
	public static final String URI = Namespaces.RDFSCHEMA + "Resource";
	
	/**
	 * Returns the identifier of the resource, either its blank node ID or its URI.
	 * 
	 * @return the resource identifier
	 */
	public String about();

	/**
	 * Returns the first property value of the property with the specified property-URI.
	 * If <code>datatypeURI</code> is null and the value is a plain literal, a string is returned.
	 * 
	 * If the property value is a plain literal and <code>datatypeURI</code> is not null,
	 * the specified datatypeURI will be applied to convert the plain literal into a Java type.
	 * 
	 * If the property value is typed literal,
	 * the specified datatype must match the datatype of the literal or an exception is thrown.
	 * 
	 * If the value is not a literal but a URI reference or a blank node, a Resource object is returned.  
	 * 
	 * @param propertyURI the property's URI
	 * @param datatypeURI the datatype's URI or null
	 * @return the property value or null if there is no statement that relates this resource with another by the specified property
	 */
	public Object getProperty(String propertyURI, String datatypeURI);
	
	/**
	 * Returns the literal property with the specified property-URI and locale. If the locale is null,
	 * the first matching literal property is returned. If locale is not null and there is no literal with 
	 * a matching <code>xml:lang</code> attribute, <code>null</code> is returned.
	 * 
	 * @param propertyURI the property's URI
	 * @param locale the locale or null
	 * @return a matching literal property or null
	 */
	public String getProperty(String propertyURI, Locale locale);
	

	/**
	 * Returns the first property with the specified property-URI.
	 * 
	 * @param propertyURI  the property's URI
	 * @return the value as Resource
	 */
	public Resource getProperty(String propertyURI);
	
	/**
	 * Returns the values of the properties with the specified property-URI as an array of resources.
	 * If there are no property values the array has length 0 (i.e. the result is never null).
	 * 
	 * @param propertyURI the property's URI
	 * @return the values as array of Resource objects
	 */
	public Resource[] getProperties(String propertyURI);
}
