/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.util;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;

/**
 * Extension of <code>java.util.Properties</code>. 
 * 
 * @author Michael Hedenus
 */
public class Properties extends java.util.Properties
{
	private static final long serialVersionUID = -1591623287823529915L;

	public Properties()
	{
		super();
	}
	
	public Properties(java.util.Properties parentProperties)
	{
		super(parentProperties);
	}

	public void load(URL url) throws IOException
	{
		InputStream inputStream = null;
		try
		{
			inputStream = url.openStream();
			this.load(inputStream);
		}
		finally
		{
			if (inputStream != null)
			{
				inputStream.close();
			}
		}
	}

	/**
	 * The value is considered true if it equals "true", "yes", "on" (all ignoring case) or "1".
	 */
	public boolean getPropertyBoolean(String key)
	{
		String val = getProperty(key);
		if (val == null)
		{
			return false;
		}
		val = val.trim();
		return "true".equalsIgnoreCase(val) || "yes".equalsIgnoreCase(val) || "on".equalsIgnoreCase(val) || "1".equals(val);
	}

	public int getPropertyInt(String key)
	{
		return Integer.parseInt(getProperty(key));
	}

	public long getPropertyLong(String key)
	{
		return Long.parseLong(getProperty(key));
	}

	public float getPropertyFloat(String key)
	{
		return Float.parseFloat(getProperty(key));
	}

	public double getPropertyDouble(String key)
	{
		return Double.parseDouble(getProperty(key));
	}

	public String getProperty(String key)
	{
		String val = super.getProperty(key);
		return val == null ? val : resolveProperty(key, val);
	}

	/**
	 * Resolves a property value by expanding references to other keys
	 * using the Ant syntax <code>${property}</code>.
	 * If a referenced property does not exist, an <code>UndefinedPropertyException</code> is thrown.
	 */
	protected String resolveProperty(String key, String value)
	{
		//System.out.println("Properties: resolving " + value);

		final StringBuffer result = new StringBuffer();
		final StringBuffer ref = new StringBuffer();
		final int len = value.length();

		for (int i = 0; i < len; i++)
		{
			char c = value.charAt(i);
			//
			// detect reference using ${x} syntax
			//
			if (c == '$' && i < (len - 2))
			{
				c = value.charAt(i + 1);
				if (c == '{')
				{
					//
					// parse property ref name
					//
					boolean refComplete = false;
					ref.setLength(0);
					for (i = i + 2; i < len; i++)
					{
						c = value.charAt(i);
						if (c == '}')
						{
							refComplete = true;
							break;
						}
						else
						{
							ref.append(c);
						}
					}
					final String refKey = ref.toString();
					if (refComplete)
					{
						// get unresolved value
						String refVal = super.getProperty(refKey);
						if (refVal == null)
						{
							//referenced but missing properties
							//are reported as errors
							throw new UndefinedPropertyException(refKey);
						}
						// recurse
						result.append(resolveProperty(key, refVal));
					}
					else
					{
						result.append("${").append(refKey);
					}
				}
				else
				{
					result.append('$');
				}
			}
			else
			{
				result.append(c);
			}
		}

		String resolvedValue = result.toString();
		//System.out.println("Properties: "+key+" = "+resolvedValue);
		return resolvedValue;
	}
}
