/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.www;

import java.io.Serializable;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * An (immuatable) internet content type (MIME). See <a href="http://tools.ietf.org/html/rfc1521">RFC 1521</a>.
 * 
 * @author Michael Hedenus
 */
public class ContentType implements Serializable, Comparable
{
	private static final long serialVersionUID = -1219725238052146375L;

	/**
	 * The default type is application/octet-stream.
	 */
	public static final ContentType DEFAULT_CONTENT_TYPE = new ContentType("application/octet-stream", "Unknown content type");

	/**
	 * A user-friendly display string.
	 */
	private final String title;

	/**
	 * The media-type.
	 */
	private final String mediaType;

	/**
	 * The sub-type.
	 */
	private final String subType;

	/**
	 * The type = media-type '/' sub-type.
	 */
	private final String type;

	/**
	 * The parameters.
	 */
	private final HashMap parameters;

	/**
	 * The complete (normalized) string.
	 */
	private String typeString;

	/**
	 * Creates a new content type.
	 * <br><b>Note</b>: the parameter keys should be lowercase. 
	 * @param mediaType
	 * @param subType
	 * @param parameters
	 */
	public ContentType(String mediaType, String subType, Map parameters)
	{
		this(mediaType, subType, parameters, null);
	}

	/**
	 * Creates a new content type.
	 * <br><b>Note</b>: the parameter keys should be lowercase.
	 * @param mediaType
	 * @param subType
	 * @param parameters
	 * @param title
	 */
	public ContentType(String mediaType, String subType, Map parameters, String title)
	{
		this.mediaType = mediaType.toLowerCase();
		this.subType = subType.toLowerCase();
		this.type = this.mediaType + "/" + this.subType;
		if (parameters != null)
		{
			this.parameters = new HashMap(parameters);
		}
		else
		{
			this.parameters = new HashMap(0);
		}

		this.title = title;
		//this.typeString = createTypeString();
	}

	/**
	 * Creates a new content type.
	 * @param spec
	 */
	public ContentType(String spec)
	{
		this(spec, null);
	}

	/**
	 * Creates a new content type.
	 * @param spec
	 * @param title
	 */
	public ContentType(String spec, String title)
	{
		StringTokenizer tokens = new StringTokenizer(spec, ";");

		//
		// type/subtype (always lowercase)
		//
		String typeDef = tokens.nextToken().trim().toLowerCase();
		int delim = typeDef.indexOf('/');
		mediaType = typeDef.substring(0, delim);
		subType = typeDef.substring(delim + 1);
		type = mediaType + "/" + subType;

		//
		// parameter
		//
		this.parameters = new HashMap(tokens.countTokens());
		while (tokens.hasMoreTokens())
		{
			String parameter = tokens.nextToken();
			delim = parameter.indexOf('=');
			String attribute = parameter.substring(0, delim).trim().toLowerCase();

			// note: parameter value is case sensitive!
			String value = parameter.substring(delim + 1).trim();

			if ((value.startsWith("\"") && value.endsWith("\""))
			/*|| (value.startsWith("'") && value.endsWith("'"))*/)
			{
				value = value.substring(1, value.length() - 1);
			}

			parameters.put(attribute, value);
		}

		this.title = title;
		//typeString = createTypeString();
	}

	private String createTypeString()
	{
		if (parameters.isEmpty())
		{
			return type;
		}
		else
		{
			final StringBuffer b = new StringBuffer().append(type);
			final String[] keys = getAttributes();
			Arrays.sort(keys);
			for (int i = 0; i < keys.length; i++)
			{
				final String key = keys[i];
				b.append(";").append(key).append("=");
				String value = (String)parameters.get(key);
				if (isToken(value))
				{
					b.append(value);
				}
				else
				{
					b.append('\"').append(value).append('\"');
				}
			}
			return b.toString();
		}
	}

	//	token  :=  1*<any (ASCII) CHAR except SPACE, CTLs,
	//   or tspecials>
	//
	//	tspecials :=  "(" / ")" / "<" / ">" / "@"
	//	 /  "," / ";" / ":" / "\" / <">
	//	 /  "/" / "[" / "]" / "?" / "="
	//	; Must be in quoted-string,
	//	; to use within parameter values
	//
	// http://tools.ietf.org/html/rfc822:
	//
	// CTL = <any ASCII control character and DEL> = 0-31 / 127
	//
	// SPACE = 32

	private static final char[] tspecials = { '(', ')', '<', '>', '@', ',', ';', ':', '\\', '\"', '/', '[', ']', '?', '=' };

	private static final boolean isToken(String s)
	{
		for (int i = 0; i < s.length(); i++)
		{
			int c = s.charAt(i);

			//
			// 1. check CTL and whitespace
			// in addition we check if for ASCII
			//
			if (c == 32 || c <= 31 || c == 127 || c > 255)
			{
				return false;
			}

			//
			// check tspecials
			//
			for (int t = 0; t < tspecials.length; t++)
			{
				if (c == tspecials[t])
				{
					return false;
				}
			}
		}

		// no illegal character found: token
		return true;
	}

	/**
	 * The full type, e.g. "text/plain".
	 */
	public String getType()
	{
		return type;
	}

	/**
	 * The media type, e.g. "text" of "text/plain".
	 */
	public String getMediaType()
	{
		return mediaType;
	}

	/**
	 * The sub type, e.g. "plain" of "text/plain".
	 */
	public String getSubType()
	{
		return subType;
	}

	/**
	 * The complete type string, e.g. "text/plain;charset=US-ASCII".
	 */
	public String getTypeString()
	{
		// optimization: lazily create the type string
		if (typeString == null)
		{
			typeString = createTypeString();
		}
		return typeString;
	}

	/**
	 * A human readable name for the type.
	 */
	public String getTitle()
	{
		return title;
	}

	/**
	 * Returns the value of parameter with the specified attribute key,. e.g. the key 'charset' may return 'US-ASCII'.
	 * <br><b>Note:</b> the key should belowercase. 
	 */
	public String getParameter(String key)
	{
		return (String)parameters.get(key);
	}

	/**
	 * Returns all attributes (parameter keys).
	 */
	public String[] getAttributes()
	{
		Set keySet = parameters.keySet();
		return (String[])keySet.toArray(new String[keySet.size()]);
	}

	public String toString()
	{
		if (title == null || "".equals(title))
		{
			return getTypeString();
		}
		else
		{
			return title + " (" + getTypeString() + ")";
		}
	}

	/**
	 * Returns true iff media-type and sub-type of this and the specified type are equal.
	 */
	public boolean typeEquals(ContentType that)
	{
		return this.getType().equals(that.getType());
	}

	/**
	 * Returns true iff media-type and sub-type of this and the specified type string are equal.
	 */
	public boolean typeEquals(String s)
	{
		return this.getType().equals(parseType(s));
	}

	/**
	 * Parses the content type of a type string, i.e. media-type/sub-type without parameters.
	 */
	public static String parseType(String s)
	{
		//extract type (chop parameters)
		int delim = s.indexOf(';');
		if (delim >= 0)
		{
			s = s.substring(0, delim).trim();
		}
		
		// type is not case sensitive -> lowercase
		return s.toLowerCase();
	}
	
	public int hashCode()
	{
		return getTypeString().hashCode();
	}

	/**
	 * Two content types are equal iff their type strings are equal.
	 */
	public boolean equals(Object obj)
	{
		if (obj != null && obj instanceof ContentType)
		{
			return this.getTypeString().equals(((ContentType)obj).getTypeString());
		}
		return false;
	}

	/**
	 * Content types are ordered by their type string.
	 */
	public int compareTo(Object obj)
	{
		return this.getTypeString().compareTo(((ContentType)obj).getTypeString());
	}

	/*
	public static void main(String[] args)
	{
		ContentType t1 = new ContentType("text/Plain     ;  Encoding=\"zip it\" ;C=\"z z\" ;charSet = ISO-8859-1; a=\"x?\"; b=y ", "Text");

		Properties params = new Properties();
		params.setProperty("charset", "ASCII-7");
		ContentType t2 = new ContentType("TEXT", "plain", params);

		System.out.println(t1);
		System.out.println(t2);

		System.out.println(t2.getAttributes().length);

		System.out.println(t1.typeEquals("TEXT/plain  	; charset"));
	}
	*/
}
