/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.www;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;

/**
 * Defines a namespace context for URIs.
 * 
 * @author Michael Hedenus
 */
public class NamespaceContext implements Serializable
{
	public static class Mapping implements Comparable, Serializable
	{
		private static final long serialVersionUID = 3544827077509909209L;

		public final String prefix;
		public final String namespaceURI;

		public Mapping(String prefix, String namespaceURI)
		{
			this.prefix = prefix;
			this.namespaceURI = namespaceURI;
		}

		public int compareTo(Object that)
		{
			return -this.namespaceURI.compareToIgnoreCase(((Mapping)that).namespaceURI);
		}

		public String toString()
		{
			return prefix + "=" + namespaceURI;
		}
	}

	private static final long serialVersionUID = 6223274343696223240L;

	private final HashMap prefix2ns = new HashMap();
	private final HashMap ns2prefix = new HashMap();
	private final ArrayList mappings = new ArrayList();
	private int nscounter = 0;

	public NamespaceContext(boolean useDefaults)
	{
		if (useDefaults)
		{
			defineNamespacePrefix("rdf", Namespaces.RDF);
			defineNamespacePrefix("rdfs", Namespaces.RDFSCHEMA);
			defineNamespacePrefix("owl", Namespaces.OWL);
			defineNamespacePrefix("dc", Namespaces.DUBLINCORE11);
			defineNamespacePrefix("xsd", Namespaces.XMLSCHEMA);
			defineNamespacePrefix("geo", Namespaces.WGS84_POS);
			defineNamespacePrefix("rss", Namespaces.RSS10);
			defineNamespacePrefix("mh", Namespaces.HEDENUS_ONTOLOGIE);
		}
	}

	/**
	 * Defines a new namespace-prefix mapping.
	 * @param prefix
	 * @param namespaceURI
	 */
	public void defineNamespacePrefix(String prefix, String namespaceURI)
	{
		if (prefix == null)
		{
			throw new IllegalArgumentException("prefix must not be null");
		}
		if (namespaceURI == null)
		{
			throw new IllegalArgumentException("namespaceURI must not be null");
		}

		Mapping mapping = new Mapping(prefix, namespaceURI);

		// put mapping in an top-down ordered list
		int i = -Collections.binarySearch(mappings, mapping) - 1;
		mappings.add(i, mapping);

		// add to indices
		prefix2ns.put(prefix, mapping);
		ns2prefix.put(namespaceURI, mapping);
	}

	/**
	 * Creates a copy of this namespace context.
	 */
	public NamespaceContext copy()
	{
		NamespaceContext copy = new NamespaceContext(false);

		for (Iterator i = mappings.iterator(); i.hasNext();)
		{
			Mapping mapping = (Mapping)i.next();
			copy.defineNamespacePrefix(mapping.prefix, mapping.namespaceURI);
		}
		return copy;
	}

	/**
	 * Returns the namespace URI for a prefix or <code>null</code> if <code>prefix</code> is not defined. 
	 */
	public String getNamespaceURI(String prefix)
	{
		if (prefix == null)
		{
			throw new IllegalArgumentException("prefix must not be null");
		}

		Mapping mapping = (Mapping)prefix2ns.get(prefix);
		return mapping == null ? null : mapping.namespaceURI;
	}

	/**
	 * Returns the prefix for a namespace URI or <code>null</code> if <code>namespaceURI</code> is not defined. 
	 */
	public String getPrefix(String namespaceURI)
	{
		if (namespaceURI == null)
		{
			throw new IllegalArgumentException("namespaceURI must not be null");
		}

		Mapping mapping = (Mapping)ns2prefix.get(namespaceURI);
		return mapping == null ? null : mapping.prefix;
	}

	/**
	 * Returns all prefixes as iterator.
	 */
	public Iterator getPrefixes()
	{
		return prefix2ns.keySet().iterator();
	}

	/**
	 * Returns all namespace URIs as iterator.
	 */
	public Iterator getNamespaces()
	{
		return ns2prefix.keySet().iterator();
	}

	/**
	 * Creates a qualified name for the given local-name and the local-name's namespace URI. 
	 * @param localName
	 * @param namespaceURI
	 */
	public String createQName(String localName, String namespaceURI)
	{
		if (localName == null)
		{
			throw new IllegalArgumentException("localName must not be null");
		}
		if (namespaceURI == null)
		{
			throw new IllegalArgumentException("namespaceURI must not be null");
		}

		if ("".equals(namespaceURI))
		{
			return localName;
		}

		String prefix = getPrefix(namespaceURI);
		if (prefix == null)
		{
			return "{" + namespaceURI + "}" + localName;
		}
		else
		{
			return prefix + ":" + localName;
		}
	}

	/**
	 * Returns the first namespace URI that match the specified URI.
	 */
	private Mapping findMapping(String uri)
	{
		final int len = mappings.size();
		for (int i = 0; i < len; i++)
		{
			Mapping mapping = (Mapping)mappings.get(i);
			if (uri.startsWith(mapping.namespaceURI))
			{
				return mapping;
			}
		}
		return null;
	}

	/**
	 * Creates an abbreviated form of the URI or returns the URI itself (same object) if no appropriate namespace mapping is found.
	 */
	public String toShortForm(String uri)
	{
		Mapping mapping = findMapping(uri);
		if (mapping != null)
		{
			return mapping.prefix + ":" + uri.substring(mapping.namespaceURI.length());
		}
		else
		{
			return uri;
		}
	}

	/**
	 * Heuristically creates a prefix and defines a new namespace-prefix mapping for the given URI.
	 * @return the generated prefix
	 */
	public String createAndDefineNamespacePrefix(String uri)
	{
		// heuristically find namespace URI
		int sep = 0;
		for (int i = uri.length() - 1; i >= 0; i--)
		{
			char c = uri.charAt(i);
			if (!Character.isLetterOrDigit(c))
			{
				sep = i;
				break;
			}
		}
		String namespaceURI = uri.substring(0, sep + 1);

		// is there already a prefix for the namespace?
		String prefix = getPrefix(namespaceURI);
		if (prefix == null)
		{
			// no prefix found for the namespace URI:
			// auto-create a prefix
			while (true)
			{
				prefix = "ns" + (++nscounter);
				if (!prefix2ns.containsKey(prefix)) /* take care that this ns does not exist */
				{
					defineNamespacePrefix(prefix, namespaceURI);
					break;
				}
			}
		}

		return prefix;
	}

	/*
	public static void main(String[] args)
	{
		NamespaceContext context = new NamespaceContext(true).copy();

		System.out.println(context.toShortForm("http://www.hedenus.de/Ontologie/Michael"));
		System.out.println(context.toShortForm("http://www.hedenus.de/Ontologie/Doris"));
		System.out.println(context.toShortForm("http://www.hedenus.de/Ontologie/"));
		System.out.println(context.toShortForm("http://www.hedenus.de/Ontologie"));

		System.out.println(context.createQName("Michael", "http://www.hedenus.de/Ontologie/"));

		//context.defineNamespacePrefix("ns2", "my:ns:2");

		String uri1 = "http://example.org/vocab/Res1";
		String uri2 = "http://example.org/vocab#Res2";
		String prefix1 = context.createAndDefineNamespacePrefix(uri1);
		String prefix2 = context.createAndDefineNamespacePrefix(uri2);

		System.out.println(context.toShortForm(uri1));
		System.out.println(context.toShortForm(uri2));

		String uri3 = "http://www.hedenus.de/Ontologie/Res3";
		String prefix3 = context.createAndDefineNamespacePrefix(uri3);
		System.out.println(context.toShortForm(uri3));

		System.out.println(context.mappings);

	}
	*/
}
