/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.www;

import java.io.File;
import java.io.IOException;

/**
 * URI utilities.
 * 
 * @author Michael Hedenus
 */
public class URIUtil
{
	private URIUtil()
	{
		//NOP
	}

	/**
	 * Resolves the relative URI against the base URI. If <code>relURI</code> is absolute, it is simply returned.
	 * @param relURI the URI to be resolved
	 * @param baseURI the base URI against which the resolving is done
	 */
	public static String resolve(String relURI, String baseURI)
	{
		String result;

		if (baseURI == null || "".equals(baseURI))
		{
			result = relURI;
		}
		else
		{
			if (relURI == null)
			{
				throw new IllegalArgumentException("URI must not be null");
			}
			if ("".equals(relURI))
			{
				result = baseURI;
			}
			else if (isAbsolute(relURI))
			{
				result = relURI;
			}
			else
			{
				int basePathBegin = getPathBegin(baseURI);
				boolean isAbsolutePath = relURI.charAt(0) == '/';
				if (isAbsolutePath)
				{
					if (basePathBegin < 0)
					{
						// "http://www.hedenus.de" + "/home/index.html" => "http://www.hedenus.de/home/index.html"
						result = baseURI + relURI;
					}
					else
					{
						// "http://www.hedenus.de/root" + "/home/index.html" => "http://www.hedenus.de/home/index.html"
						result = baseURI.substring(0, basePathBegin) + relURI;
					}
				}
				else
				{
					if (basePathBegin < 0)
					{
						// "http://www.hedenus.de" + "home/index.html" = > "http://www.hedenus.de/home/index.html"
						result = baseURI + "/" + relURI;
					}
					else
					{
						// resolve relative to last "/", i.e. exclude file
						int last = baseURI.lastIndexOf('/');
						result = baseURI.substring(0, last + 1) + relURI;
					}
				}
			}
		}
		return result;
	}

	/**
	 * Returns the index where the path of the URI begins or -1 if
	 * no path component is detected.
	 * If the URI is not absolute the path is identical with the URI itself.
	 * <br><b>Note:</b> If e.g. the URI is <code>http://www.hedenus.de</code> the path index
	 * is -1 but for  <code>http://www.hedenus.de/</code> the index is 21.
	 */
	private static int getPathBegin(String uri)
	{
		int pi;
		if (isAbsolute(uri))
		{
			// check for authority
			int ai = uri.indexOf("//");
			if (ai >= 0)
			{
				pi = uri.indexOf('/', ai + 2);
			}
			else
			{
				pi = uri.indexOf('/');
			}
		}
		else
		{
			// if not absolute consider it as relative ==> path = uri
			pi = 0;
		}
		return pi;
	}

	/**
	 * Gets path of the URI or <code>null</code> if no path is detected.
	 */
	public static String getPath(String uri)
	{
		int pi = getPathBegin(uri);
		if (pi < 0)
		{
			return null;
		}
		else
		{
			return uri.substring(pi);
		}
	}

	/**
	 * Tests wether the URI starts with a legal scheme, e.g. <code>http:</code>.
	 */
	public static boolean isAbsolute(String uri)
	{
		return getSchemeEnd(uri) > 0;
	}

	/**
	 * Returns the URI's scheme nor <code>null</code> if the URI has no legal scheme definition.
	 */
	public static String getScheme(String uri)
	{
		int i = getSchemeEnd(uri);
		if (i <= 0)
		{
			return null;
		}
		else
		{
			return uri.substring(0, i);
		}
	}

	/**
	 * Returns the end index of the scheme part.
	 */
	private static int getSchemeEnd(String uri)
	{
		final int len = uri.length();
		for (int i = 0; i < len; i++)
		{
			char c = uri.charAt(i);
			if (c == ':')
			{
				return i;
			}
			else if (!isSchemeChar(c, i))
			{
				return -1;
			}
		}
		return -1;
	}

	/**
	 * Translates a <code>file:</code> URI to a file.
	 * The host name will be ignored.
	 */
	public static File toFile(String uri)
	{
		if (uri == null)
		{
			throw new IllegalArgumentException("URI must not be null");
		}

		if (uri.startsWith("file://"))
		{
			int hosti = uri.indexOf('/', 7);
			//String host = uri.substring(7, hosti);
			int pathi = hosti + 1;

			StringBuffer file = new StringBuffer();
			for (int i = pathi; i < uri.length(); i++)
			{
				char c = uri.charAt(i);
				switch (c)
				{
					case '%':
						int high = hexDigitToInt(uri.charAt(i + 1));
						int low = hexDigitToInt(uri.charAt(i + 2));
						int code = high * 16 + low;
						file.append((char)code);
						i += 2;
						break;
					default:
						file.append(c);
						break;
				}
			}

			return new File(file.toString());
		}
		else
		{
			throw new URIException("not a 'file:' URI");
		}
	}

	/**
	 * Translates a file to a <code>file:</code> URI with no specified host.
	 */
	public static String toURI(File file)
	{
		return toURI(file, null);
	}

	/**
	 * Translates a file to a <code>file:</code> URI using the specified host name.
	 */
	public static String toURI(File file, String host)
	{
		if (file == null)
		{
			throw new IllegalArgumentException("file must not be null");
		}

		try
		{
			String path = file.getCanonicalPath();
			StringBuffer buffer = new StringBuffer();
			buffer.append("file://");
			if (host != null)
			{
				buffer.append(host);
			}
			buffer.append("/");

			final int start = path.charAt(0) == '/' ? 1 : 0;
			for (int i = start; i < path.length(); i++)
			{
				char c = path.charAt(i);
				if (c == fileSepChar)
				{
					buffer.append('/');
				}
				else if (isLegalChar(c))
				{
					buffer.append(c);
				}
				else
				{
					buffer.append('%').append(Integer.toHexString(c));
				}
			}

			return buffer.toString();
		}
		catch (IOException e)
		{
			throw new URIException("cannot create canonical path: " + e.getMessage());
		}
	}

	private static final char fileSepChar = System.getProperty("file.separator").charAt(0);

	/**
	 * Returns true if the char is legal character for the scheme.
	 * @param c the char
	 * @param index the char's position in the URI string
	 */
	public static boolean isSchemeChar(char c, int index)
	{
		if (index == 0)
		{
			return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
		}
		else
		{
			return isLetterOrDigit(c) || c == '+' || c == '-' || c == '.';
		}
	}

	/**
	 * Returns true if the character is a letter or number.
	 */
	public static boolean isLetterOrDigit(char c)
	{
		return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9');
	}

	/**
	 * Returns true if the character is a letter or digit or 'other legal character'.
	 */
	public static boolean isLegalChar(char c)
	{
		return isLetterOrDigit(c) || isOtherLegalChar(c);
	}

	/**
	 * Returns true if the character is 'other legal character'.
	 */
	public static boolean isOtherLegalChar(char c)
	{
		final int otherLegalCharCount = otherLegalChars.length();
		for (int i = 0; i < otherLegalCharCount; i++)
		{
			if (otherLegalChars.charAt(i) == c)
			{
				return true;
			}
		}
		return false;
	}

	private static final String otherLegalChars = "_-!.~'()*,;:$&+=?/[]@";

	private static int hexDigitToInt(char c)
	{
		if (c >= '0' && c <= '9')
		{
			return c - '0';
		}
		if (c >= 'a' && c <= 'f')
		{
			return c - 'a' + 10;
		}
		if (c >= 'A' && c <= 'F')
		{
			return c - 'A' + 10;
		}
		throw new URIException("illegal character in hex code: " + c);
	}

	public static void main(String[] args)
	{
		//File file = new File("C:\\Documents and Settings\\uid10771\\Desktop\\Verknpfung mit eclipse.exe.lnk");
		//String fileURI = toURI(file);
		//System.out.println(fileURI);
		//System.out.println(toFile(fileURI));

		/*
		 System.out.println(getScheme("http://www."));
		 System.out.println(getScheme(":"));
		 System.out.println(getScheme("a:"));
		 System.out.println(getScheme("urn:my:self"));
		 System.out.println(getScheme("~:mm"));

		 System.out.println(getPathBegin("http://www.hedenus.de"));
		 System.out.println(getPathBegin("http://www.hedenus.de/"));
		 System.out.println(getPathBegin(""));
		 System.out.println(getPathBegin("/my"));
		 */
	}
}
