/*****************************************************************************
 *  Copyright 2008 Michael Hedenus                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *****************************************************************************/
package de.hedenus.xmltypes;

/**
 * An interface describing a XML type (for example W3 Schema or other).
 * 
 * @author Michael Hedenus
 */
public interface Type
{	
	/**
	 * Returns the mapped Java type.
	 */
	public Class getRepresentationClass();
	
	/**
	 * Returns true iff the value is valid.
	 */
	public boolean validate(Object value);
	
	/**
	 * Returns the base types of the type.
	 * <ul>
	 * <li>If the type is a union type the constituent types are returned.</li>
	 * <li>If the type is a constrained type the unconstrained type is returned.</li>
	 * <li>If the type is a unconstrained type the base types in the type hierarchy are returned.</li>
	 * <li>If the type is a root of the type hierarchy, <code>null</code> is returned.
	 * <ul>
	 */
	public Type[] getBaseTypes();
	
	/**
	 * Returns true iff the type is considered to be a primitive.
	 * The definition of a primitive type depends on the type system. Note that a constrained type
	 * may be considered as a primitive,
	 * e.g. the boolean primitive <i>can</i> be defined as an element of the set {0;1}, which
	 * is a constraint of the integral numbers.
	 * On the other hand, the boolean primitive <i>can also</i> be defined
	 * as in instance of {true;false} which is not a constraint of any other set.
	 * <br>In terms of W3 Schema primitive datatypes are those that are not defined in terms of other datatypes.
	 */
	public boolean isPrimitive();
	
	/**
	 * Returns true iff the type is abstract.
	 * If a type is abstract the methods toLiteral() and parse() will throw an exception.
	 * <br>In terms of W3 Schema this is an <i>ur-type</i>. 
	 */
	public boolean isAbstract();
	
	/**
	 * Returns the name of the type according to the type system, for example &quot;http://www.w3.org/2001/XMLSchema#string&quot;.
	 */
	public String getName();
	
	/**
	 * Returns the type system identifier, for example &quot;http://www.w3.org/2001/XMLSchema&quot;
	 */
	public String getTypeSystem();
	
	/**
	 * Returns a string representation of the type, for example &quot;http://www.w3.org/2001/XMLSchema#string&quot;
	 */
	public String toString();
	
	/**
	 * Returns a lexical representation (literal) of a value.
	 * This method <i>assumes</i> that the value is valid, so before
	 * passing a value to this method it should be checked if it is really appropriate.
	 */
	public String toLiteral(Object value) throws TypeException;
	
	/**
	 * Parses a lexical value. Only the unconstrained base types may be used for parsing, i.e.
	 * it is possible that the parsing is successful but the resulting object is not valid
	 * according to the constraints of this particluar type.
	 * For example let the type be http://www.w3.org/2001/XMLSchema#nonNegativeInteger and the value be -5 then
	 * it is up to the implementation either to throw an exception (because the value is negative)
	 * or to return the value and let the application check it using the validate() method.
	 * The latter case would allow the user to be informed and to correct the value.
	 * <br>If the method is successful then the returned value should at least be valid according to
	 * the unconstrained base type of this particalur type.
	 */
	public Object parse(String literal) throws TypeException;
}